# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List

import pytest
import torch

from gluonts.torch.util import lagged_sequence_values


@pytest.mark.parametrize(
    "lag_indices, prior_sequence, sequence",
    [
        (
            [0, 1, 5, 10, 20],
            torch.randn((4, 100)),
            torch.randn((4, 8)),
        ),
        (
            [0, 1, 5, 10, 20],
            torch.randn((4, 100, 1)),
            torch.randn((4, 8, 1)),
        ),
        (
            [0, 1, 5, 10, 20],
            torch.randn((4, 100, 2)),
            torch.randn((4, 8, 2)),
        ),
    ],
)
def test_lagged_sequence_values(
    lag_indices: List[int],
    prior_sequence: torch.Tensor,
    sequence: torch.Tensor,
):
    res = lagged_sequence_values(lag_indices, prior_sequence, sequence)
    full_sequence = torch.cat((prior_sequence, sequence), dim=1)
    for t in range(res.shape[1]):
        expected_lags_t = torch.stack(
            [
                full_sequence[:, t + prior_sequence.shape[1] - l]
                for l in lag_indices
            ],
            dim=-1,
        ).reshape(sequence.shape[0], -1)
        assert torch.allclose(expected_lags_t, res[:, t, :])



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from __future__ import annotations
from typing import Any, List, Optional
import numpy as np
import numpy.typing as npt
import pandas as pd
from sklearn.base import TransformerMixin
from tsbench.evaluations.metrics import Performance


class PerformanceTransformer(TransformerMixin):
    """
    The performance transformer transforms performances into model outputs for
    supervised learning as well as model outputs to performance objects.
    """

    def __init__(self, metrics: list[str] | None = None):
        """
        Args:
            metrics: The performance metrics to transform. If `None`, all performance metrics are
                transformed. If provided, metrics that are not present will be converted into NaNs.
        """
        self.encoder = PerformanceEncoder(metrics)

    @property
    def features_names_(self) -> list[str]:
        """
        Returns the feature names for the columns of the transformed
        performance objects.
        """
        return self.encoder.feature_names_

    def fit(self, y: list[Performance]) -> PerformanceTransformer:
        """
        Uses the provided performances to fit the performance transformer.

        Args:
            y: The performance objects.
        """
        self.encoder.fit(y)
        return self

    def transform(self, y: list[Performance]) -> npt.NDArray[np.float32]:
        """
        Transforms the provided performance object into NumPy arrays according
        to the fitted transformer.

        Args:
            y: The performance objects.

        Returns:
            An array of shape [N, K] of transformed performance objects (N: the number of
                performance objects, K: number of performance metrics).
        """
        return self.encoder.transform(y)

    def inverse_transform(
        self, y: npt.NDArray[np.float32]
    ) -> list[Performance]:
        """
        Transforms the provided NumPy arrays back into performance objects
        according to the fitted transformer.

        Args:
            y: A NumPy array of shape [N, K] of performances (N: number of performances, K: number
                of performance metrics).

        Returns:
            The performance objects.
        """
        return self.encoder.inverse_transform(y)


# -------------------------------------------------------------------------------------------------
# pylint: disable=missing-class-docstring,missing-function-docstring


class PerformanceEncoder:
    def __init__(self, metrics: list[str] | None = None):
        self.metrics = metrics
        self.all_feature_names_: list[str]
        self.feature_names_: list[str]

    def fit(self, X: list[Performance], _y: Any = None) -> PerformanceEncoder:
        df = Performance.to_dataframe(X)
        self.all_feature_names_ = df.columns.tolist()
        if self.metrics is None:
            self.feature_names_ = df.columns.tolist()
        else:
            assert all(m in df.columns for m in self.metrics)
            self.feature_names_ = self.metrics
        return self

    def transform(
        self, X: list[Performance], _y: Any = None
    ) -> npt.NDArray[np.float32]:
        df = Performance.to_dataframe(X)
        return df[self.feature_names_].to_numpy()

    def inverse_transform(
        self, X: npt.NDArray[np.float32], _y: Any = None
    ) -> list[Performance]:
        df = pd.DataFrame(X, columns=self.feature_names_).assign(
            **{
                col: np.nan
                for col in set(self.all_feature_names_)
                - set(self.feature_names_)
            }
        )
        return [
            Performance.from_dict(row.to_dict()) for _, row in df.iterrows()
        ]



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

"""
This example show how to intialize the network with parameters from a model that was previously trained.
"""

from gluonts.dataset.repository.datasets import get_dataset, dataset_recipes
from gluonts.evaluation import backtest_metrics, Evaluator
from gluonts.model.simple_feedforward import SimpleFeedForwardEstimator
from gluonts.mx.trainer import Trainer
import pandas as pd


if __name__ == "__main__":
    print(f"datasets available: {dataset_recipes.keys()}")

    # we pick m4_hourly as it only contains a few hundred time series
    dataset = get_dataset("m4_hourly", regenerate=False)

    # First train a model
    estimator = SimpleFeedForwardEstimator(
        prediction_length=dataset.metadata.prediction_length,
        trainer=Trainer(epochs=10, num_batches_per_epoch=10),
    )

    train1_output = estimator.train_model(dataset.train)

    # callback to overwrite parameters of the new model with the already trained model
    def copy_params(net):
        params1 = train1_output.trained_net.collect_params()
        params2 = net.collect_params()
        for p1, p2 in zip(params1.values(), params2.values()):
            p2.set_data(p1.data())

    estimator = SimpleFeedForwardEstimator(
        prediction_length=dataset.metadata.prediction_length,
        trainer=Trainer(
            epochs=5, num_batches_per_epoch=10, post_initialize_cb=copy_params
        ),
    )

    new_pred = estimator.train(dataset.train)

    ev = Evaluator(num_workers=0)
    agg_metrics1, _ = backtest_metrics(
        dataset.test, train1_output.predictor, evaluator=ev
    )
    agg_metrics2, _ = backtest_metrics(dataset.test, new_pred, evaluator=ev)

    df = pd.DataFrame([agg_metrics1, agg_metrics2], index=["model1", "model2"])
    print(df)



# Configuration file for the Sphinx documentation builder.
#
# This file only contains a selection of the most common options. For a full
# list see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Path setup --------------------------------------------------------------

# If extensions (or modules to document with autodoc) are in another directory,
# add these directories to sys.path here. If the directory is relative to the
# documentation root, use os.path.abspath to make it absolute, like shown here.
#
import os
import sys

sys.path.insert(0, os.path.abspath("."))

from gluonts.meta import colors

# -- Project information -----------------------------------------------------

project = "GluonTS"
copyright = "2022, Amazon"
author = "Amazon"


# -- General configuration ---------------------------------------------------

# Add any Sphinx extension module names here, as strings. They can be
# extensions coming with Sphinx (named 'sphinx.ext.*') or your custom
# ones.
extensions = [
    "sphinx.ext.autodoc",
    "sphinx.ext.autosummary",
    "sphinx.ext.doctest",
    "sphinx.ext.viewcode",
    "sphinx.ext.intersphinx",
    "sphinx.ext.napoleon",
    "sphinx.ext.todo",
    "sphinx.ext.coverage",
    "sphinx.ext.mathjax",
    "sphinx.ext.ifconfig",
    "nbsphinx",
    "IPython.sphinxext.ipython_console_highlighting",
    "IPython.sphinxext.ipython_directive",
    "myst_parser",
    "mdinclude",
]

# Add any paths that contain templates here, relative to this directory.
templates_path = ["_templates"]

# List of patterns, relative to source directory, that match files and
# directories to ignore when looking for source files.
# This pattern also affects html_static_path and html_extra_path.
exclude_patterns = ["_build", "Thumbs.db", ".DS_Store"]

# The name of the Pygments (syntax highlighting) style to use.
pygments_style = "default"

# -- Options for HTML output -------------------------------------------------

# The theme to use for HTML and HTML Help pages.  See the documentation for
# a list of builtin themes.
#
html_theme = "furo"


html_theme_options = {
    "light_css_variables": {
        "color-brand-primary": colors.RED,
        "color-brand-content": colors.RED,
        "color-announcement-text": colors.RED,
        "color-announcement-background": "var(--color-background-secondary)",
    },
    "dark_css_variables": {
        "color-brand-primary": colors.GREEN,
        "color-brand-content": colors.GREEN,
        "color-announcement-text": colors.GREEN,
        "color-announcement-background": "var(--color-background-secondary)",
    },
}

if os.environ.get("GITHUB_REF_NAME") == "dev":
    html_theme_options[
        "announcement"
    ] = "<strong>Note:</strong> You are looking at the development docs."


# Add any paths that contain custom static files (such as style sheets) here,
# relative to this directory. They are copied after the builtin static files,
# so a file named "default.css" will overwrite the builtin "default.css".
html_static_path = ["_static"]

html_css_files = ["style.css"]

# The name of an image file (relative to this directory) to place at the top
# of the sidebar.
html_logo = "_static/logos/gluonts.svg"

# The name of an image file (relative to this directory) to use as a favicon of
# the docs.  This file should be a Windows icon file (.ico) being 16x16 or 32x32
# pixels large.
html_favicon = "_static/gluonts.ico"

# Enable Markdown
source_suffix = [".rst", ".md"]



import torch

import numpy as np

from gluonts.dataset.repository.datasets import get_dataset

from gluonts.evaluation import Evaluator

from gluonts.torch.model.deepar import DeepAREstimator
from gluonts.torch.distributions import (
    SplicedBinnedPareto,
    SplicedBinnedParetoOutput,
)

from matplotlib import pyplot as plt


first_batch_dim = 2
second_batch_dim = 3


bins_lower_bound, bins_upper_bound = -25.0, 25.0
nbins = 11
percentile_tail = 0.05
output_dim = 1

logits = torch.tensor(
    [1.0, 2.0, 3.0, 4.0, 3.0, 2.0, 1.0, 1.0, 1.0, 5.0, 1.0]
).unsqueeze(0)
logits = torch.tensor(
    [0.0, 2.0, 2.0, 2.0, 0.0, 0.0, 1.0, 0.0, 1.0, 3.0, 0.0]
).unsqueeze(0)
logits = logits.repeat(first_batch_dim, second_batch_dim, 1)


binned_distr = SplicedBinnedPareto(
    bins_lower_bound=bins_lower_bound,
    bins_upper_bound=bins_upper_bound,
    tail_percentile_gen_pareto=0.05,
    numb_bins=nbins,
    logits=logits,
    lower_gp_xi=torch.tensor(0.5, dtype=torch.float).repeat(
        first_batch_dim, second_batch_dim, 1
    ),
    lower_gp_beta=torch.tensor(3.5, dtype=torch.float).repeat(
        first_batch_dim, second_batch_dim, 1
    ),
    upper_gp_xi=torch.tensor(0.5, dtype=torch.float).repeat(
        first_batch_dim, second_batch_dim, 1
    ),
    upper_gp_beta=torch.tensor(1.0, dtype=torch.float).repeat(
        first_batch_dim, second_batch_dim, 1
    ),
)

## Getting the median, mean and mode
print("Median: ", binned_distr.median)
print("Mean: ", binned_distr.mean)
print("Mode: ", binned_distr.mode)


## Log prob
# print(torch.tensor([-1.0]).repeat(first_batch_dim, second_batch_dim).shape)
print(
    binned_distr.log_prob(
        torch.tensor([-1.0]).repeat(first_batch_dim, second_batch_dim)
    )
)


## ICDF:
quantiles = []

plot_range = np.linspace(0.0, 1.0, 300)
for q in plot_range:
    quantiles.append(
        binned_distr.icdf(torch.tensor([q]))[0, 0].detach().cpu().numpy()
    )
plt.plot(plot_range, quantiles)
plt.show()

## CDF:
quantiles = []

plot_range = range(-40, 30)
for q in plot_range:
    quantiles.append(
        binned_distr.cdf(torch.tensor([q]).repeat(2, 3))[0, 0]
        .detach()
        .cpu()
        .numpy()
    )
plt.plot(plot_range, quantiles)

plt.show()

## PDF:
quantiles = []

plot_range = range(-40, 35)
for q in plot_range:
    quantiles.append(
        binned_distr.pdf(torch.tensor([q]).repeat(2, 3))[0, 0]
        .detach()
        .cpu()
        .numpy()
    )
plt.plot(plot_range, quantiles)

plt.plot(binned_distr.mean[0, 0], 0.0, "o")
plt.plot(binned_distr.median[0, 0], 0.0, "o")
plt.plot(binned_distr.mode[0, 0], 0.0, "o")

plt.legend(["pdf", "mean", "median", "mode"])

plt.show()


### Empirical PDF from samples:

samples = binned_distr.sample(torch.tensor(range(0, 5000)).shape)

print("samples", samples.shape)

plt.hist(samples[:, 0, 0].detach().cpu().numpy(), bins=100)
plt.xlim([-45, 35])
plt.show()


### Fitting the distribution:

dataset = get_dataset("electricity")
evaluator = Evaluator(quantiles=[0.1, 0.5, 0.9])

prediction_length = 2  # dataset.metadata.prediction_length

print(prediction_length)

deepar_estimator_spline = DeepAREstimator(
    freq=dataset.metadata.freq,
    prediction_length=prediction_length,
    num_feat_static_cat=len(dataset.metadata.feat_static_cat),
    cardinality=[int(f.cardinality) for f in dataset.metadata.feat_static_cat],
    # distr_output=PiecewiseLinearOutput(20),
    # loss = CRPS(),
    distr_output=SplicedBinnedParetoOutput(
        bins_lower_bound=0.0,
        bins_upper_bound=100.0,
        num_bins=22,
        tail_percentile_gen_pareto=0.05,
    ),
    batch_size=2,
    trainer_kwargs={
        "gpus": -1 if torch.cuda.is_available() else None,
        "max_epochs": 5,  # Put the number of epoch here
    },
)


deepar_predictor_spline = deepar_estimator_spline.train(
    dataset.train,
    cache_data=True,
)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Any, Callable, Dict, List, Optional, Type, TypeVar
from tsbench.config import EnsembleConfig, ModelConfig
from tsbench.evaluations.tracking import EnsembleTracker, ModelTracker
from ._base import DatasetFeaturesMixin, Surrogate

SURROGATE_REGISTRY: Dict[str, Type[Surrogate[ModelConfig]]] = {}
ENSEMBLE_SURROGATE_REGISTRY: Dict[str, Type[Surrogate[EnsembleConfig]]] = {}


S = TypeVar("S", bound=Type[Surrogate[ModelConfig]])
E = TypeVar("E", bound=Type[Surrogate[EnsembleConfig]])


def register_surrogate(name: str) -> Callable[[S], S]:
    """
    Registers the provided class with the given name in the global surrogate
    registry.
    """

    def register(cls: S) -> S:
        SURROGATE_REGISTRY[name] = cls
        return cls

    return register


def register_ensemble_surrogate(name: str) -> Callable[[E], E]:
    """
    Registers the provided class with the given name in the global ensemble
    surrogate registry.
    """

    def register(cls: E) -> E:
        ENSEMBLE_SURROGATE_REGISTRY[name] = cls
        return cls

    return register


def create_surrogate(
    name: str,
    tracker: ModelTracker,
    predict: Optional[List[str]],
    input_flags: Dict[str, bool],
    **kwargs: Any,
) -> Surrogate[ModelConfig]:
    """
    Creates a surrogate using the specified parameters.
    """
    assert name in SURROGATE_REGISTRY, f"Unknown surrogate {name}."

    # Build the parameters
    surrogate_cls = SURROGATE_REGISTRY[name]
    args = {"predict": predict, "tracker": tracker, **kwargs}
    if issubclass(surrogate_cls, DatasetFeaturesMixin):
        args.update(input_flags)

    # Initialize the surrogate
    return surrogate_cls(**args)  # type: ignore


def create_ensemble_surrogate(
    name: str,
    tracker: EnsembleTracker,
    predict: Optional[List[str]],
    input_flags: Dict[str, bool],
    **kwargs: Any,
) -> Surrogate[EnsembleConfig]:
    """
    Creates an ensemble surrogate using the specified parameters.
    """
    assert name in ENSEMBLE_SURROGATE_REGISTRY, f"Unknown surrogate {name}."

    # Build the parameters
    surrogate_cls = ENSEMBLE_SURROGATE_REGISTRY[name]
    args = {"predict": predict, "tracker": tracker, **kwargs}
    if issubclass(surrogate_cls, DatasetFeaturesMixin):
        args.update(input_flags)

    # Initialize the surrogate
    return surrogate_cls(**args)  # type: ignore



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List, Tuple

import numpy as np
from numba import jit


def range_overlap(
    left_range: range,
    right_range: range,
) -> bool:
    """
    Checks if two ranges have an overlap. Here each range is assumed to be
    consecutive, i.e., `step` field of `range` is always 1.

    Parameters
    ----------
    left_range
    right_range

    Returns
    -------
    True or False depending on the overlap.
    """
    if left_range[0] <= right_range[-1] and left_range[-1] >= right_range[0]:
        return True
    return False


@jit(nopython=True)
def labels_to_ranges_numba(labels: np.ndarray) -> List[Tuple]:
    """
    Converts the given list of labels to list of anomaly (defined by positive
    label) ranges where range is represented by a pair of integers (to make
    numba work).

    Parameters
    ----------
    labels
        Boolean list of labels.

    Returns
    -------
    List of ranges.
    """

    ranges_ls = []
    start_ix = None
    stop_ix = None
    for ix, label in enumerate(labels):
        if label:
            # this might be the last positive label in the anomaly range
            stop_ix = ix + 1
            if start_ix is None:
                # a new positive label is seen
                start_ix = ix
        elif start_ix is not None:
            # a consecutive sequence of positive labels is ended
            assert stop_ix is not None
            ranges_ls.append((start_ix, stop_ix))
            start_ix = None

    if start_ix is not None:
        # the last element of `labels` is True, we never ended that range
        assert stop_ix is not None
        ranges_ls.append((start_ix, stop_ix))

    return ranges_ls


def labels_to_ranges(labels: List[bool]) -> List[range]:
    """
    Converts the given list of labels to list of anomaly (defined by positive
    label) ranges.

    Parameters
    ----------
    labels
        Boolean list of labels.

    Returns
    -------
    List of ranges.
    """

    labels_np = np.array(labels)
    labels_np[np.isnan(labels_np)] = 0
    ls_pairs = labels_to_ranges_numba(labels_np)
    return [range(pair[0], pair[1]) for pair in ls_pairs]


def ranges_to_singletons(
    ranges: List[range],
) -> List[range]:
    """
    Convenient function to convert list of consecutive ranges to list of
    singleton ranges.

    Parameters
    ----------
    ranges
        List of ranges.

    Returns
    -------
    List of singleton ranges.
    """
    assert all(
        r.step == 1 and r.start >= 0 and r.stop >= 0 for r in ranges
    ), "Ranges should be consecutive and contain non-negative indices."

    return [range(i, i + 1) for r in ranges for i in r]



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Iterator

import torch

from gluonts.model.forecast_generator import (
    recursively_zip_arrays,
    make_distribution_forecast,
)
from gluonts.torch.model.forecast import DistributionForecast


@recursively_zip_arrays.register(torch.Tensor)
def _(x: torch.Tensor) -> Iterator[torch.Tensor]:
    for i in range(x.shape[0]):
        yield x[i]


@make_distribution_forecast.register(torch.distributions.Distribution)
def _(
    distr: torch.distributions.Distribution, *args, **kwargs
) -> DistributionForecast:
    return DistributionForecast(distr, *args, **kwargs)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import os
from pathlib import Path
from typing import Dict, Optional, List, Union, Iterator

import numpy as np

from gluonts.core.component import validated
from gluonts.dataset.common import Dataset
from gluonts.dataset.util import forecast_start
from gluonts.model.forecast import SampleForecast, QuantileForecast
from gluonts.model.predictor import RepresentablePredictor
from gluonts.time_feature import get_seasonality

# https://stackoverflow.com/questions/25329955/check-if-r-is-installed-from-python
from subprocess import Popen, PIPE

proc = Popen(["which", "R"], stdout=PIPE, stderr=PIPE)
R_IS_INSTALLED = proc.wait() == 0

try:
    import rpy2.robjects.packages as rpackages
    from rpy2 import rinterface, robjects
    from rpy2.rinterface import RRuntimeError
except ImportError as e:
    rpy2_error_message = str(e)
    RPY2_IS_INSTALLED = False
else:
    RPY2_IS_INSTALLED = True

USAGE_MESSAGE = """
The RForecastPredictor is a thin wrapper for calling the R forecast package.
In order to use it you need to install R and run

pip install 'rpy2>=2.9.*,<3.*'

R -e 'install.packages(c("forecast", "nnfor"),\
repos="https://cloud.r-project.org")'
"""

SAMPLE_FORECAST_METHODS = ["ets", "arima"]
QUANTILE_FORECAST_METHODS = ["tbats", "thetaf", "stlar"]
POINT_FORECAST_METHODS = ["croston", "mlp"]
SUPPORTED_METHODS = (
    SAMPLE_FORECAST_METHODS
    + QUANTILE_FORECAST_METHODS
    + POINT_FORECAST_METHODS
)


class RForecastPredictor(RepresentablePredictor):
    """
    Wrapper for calling the `R forecast package.

    <http://pkg.robjhyndman.com/forecast/>`_.

    The `RForecastPredictor` is a thin wrapper for calling the R forecast
    package.  In order to use it you need to install R and run::

        pip install 'rpy2>=2.9.*,<3.*'
        R -e 'install.packages(c("forecast", "nnfor"), repos="https://cloud.r-project.org")'

    Parameters
    ----------
    freq
        The granularity of the time series (e.g. '1H')
    prediction_length
        Number of time points to be predicted.
    method
        The method from rforecast to be used one of
        "ets", "arima", "tbats", "croston", "mlp", "thetaf".
    period
        The period to be used (this is called `frequency` in the R forecast
        package), result to a tentative reasonable default if not specified
        (for instance 24 for hourly freq '1H')
    trunc_length
        Maximum history length to feed to the model (some models become slow
        with very long series).
    params
        Parameters to be used when calling the forecast method default.
        Note that currently only `output_type = 'samples'` is supported.
    """  # noqa: E501

    @validated()
    def __init__(
        self,
        freq: str,
        prediction_length: int,
        method_name: str = "ets",
        period: Optional[int] = None,
        trunc_length: Optional[int] = None,
        params: Optional[Dict] = None,
    ) -> None:
        super().__init__(prediction_length=prediction_length)

        if not R_IS_INSTALLED:
            raise ImportError("R is not Installed! \n " + USAGE_MESSAGE)

        if not RPY2_IS_INSTALLED:
            raise ImportError(rpy2_error_message + USAGE_MESSAGE)

        self._robjects = robjects
        self._rinterface = rinterface
        self._rinterface.initr()
        self._rpackages = rpackages

        this_dir = os.path.dirname(os.path.realpath(__file__))
        this_dir = this_dir.replace("\\", "/")  # for windows
        r_files = [
            n[:-2] for n in os.listdir(f"{this_dir}/R/") if n[-2:] == ".R"
        ]

        for n in r_files:
            try:
                path = Path(this_dir, "R", f"{n}.R")
                robjects.r(f'source("{path}")'.replace("\\", "\\\\"))
            except RRuntimeError as er:
                raise RRuntimeError(str(er) + USAGE_MESSAGE) from er

        assert method_name in SUPPORTED_METHODS, (
            f"method {method_name} is not supported please use one of"
            f" {SUPPORTED_METHODS}"
        )

        self.method_name = method_name

        self._stats_pkg = rpackages.importr("stats")
        self._r_method = robjects.r[method_name]

        self.prediction_length = prediction_length
        self.period = period if period is not None else get_seasonality(freq)
        self.trunc_length = trunc_length

        self.params = {
            "prediction_length": self.prediction_length,
            "output_types": ["samples"],
            "frequency": self.period,
        }
        if params is not None:
            self.params.update(params)

    def _unlist(self, l):
        if type(l).__name__.endswith("Vector"):
            return [self._unlist(x) for x in l]
        else:
            return l

    def _run_r_forecast(self, d, params, save_info):
        buf = []

        def save_to_buf(x):
            buf.append(x)

        def dont_save(x):
            pass

        f = save_to_buf if save_info else dont_save

        # save output from the R console in buf
        self._rinterface.set_writeconsole_regular(f)
        self._rinterface.set_writeconsole_warnerror(f)

        make_ts = self._stats_pkg.ts
        r_params = self._robjects.vectors.ListVector(params)
        vec = self._robjects.FloatVector(d["target"])
        ts = make_ts(vec, frequency=self.period)
        forecast = self._r_method(ts, r_params)
        forecast_dict = dict(
            zip(forecast.names, map(self._unlist, list(forecast)))
        )

        if "quantiles" in forecast_dict or "upper_quantiles" in forecast_dict:

            def from_interval_to_level(interval: int, side: str):
                if side == "upper":
                    level = 50 + interval / 2
                elif side == "lower":
                    level = 50 - interval / 2
                else:
                    raise ValueError
                return level / 100

            # Post-processing quantiles on then Python side for the convenience
            # of asserting and debugging.
            upper_quantiles = [
                str(from_interval_to_level(interval, side="upper"))
                for interval in params["intervals"]
            ]

            lower_quantiles = [
                str(from_interval_to_level(interval, side="lower"))
                for interval in params["intervals"]
            ]

            # Median forecasts would be available at two places: Lower 0 and
            # Higher 0 (0-prediction interval)
            forecast_dict["quantiles"] = dict(
                zip(
                    lower_quantiles + upper_quantiles[1:],
                    forecast_dict["lower_quantiles"]
                    + forecast_dict["upper_quantiles"][1:],
                )
            )

            # `QuantileForecast` allows "mean" as the key; we store them as
            # well since they can differ from median.
            forecast_dict["quantiles"].update(
                {"mean": forecast_dict.pop("mean")}
            )

        self._rinterface.set_writeconsole_regular(
            self._rinterface.consolePrint
        )
        self._rinterface.set_writeconsole_warnerror(
            self._rinterface.consolePrint
        )
        return forecast_dict, buf

    def predict(
        self,
        dataset: Dataset,
        num_samples: int = 100,
        intervals: Optional[List] = None,
        save_info: bool = False,
        **kwargs,
    ) -> Iterator[Union[SampleForecast, QuantileForecast]]:
        if self.method_name in POINT_FORECAST_METHODS:
            print(
                "Overriding `output_types` to `mean` since"
                f" {self.method_name} is a point forecast method."
            )
        elif self.method_name in QUANTILE_FORECAST_METHODS:
            print(
                "Overriding `output_types` to `quantiles` since "
                f"{self.method_name} is a quantile forecast method."
            )

        for data in dataset:
            if self.trunc_length:
                shift_by = max(data["target"].shape[0] - self.trunc_length, 0)
                data["start"] = data["start"] + shift_by
                data["target"] = data["target"][-self.trunc_length :]

            params = self.params.copy()
            params["num_samples"] = num_samples

            if self.method_name in POINT_FORECAST_METHODS:
                params["output_types"] = ["mean"]
            elif self.method_name in QUANTILE_FORECAST_METHODS:
                params["output_types"] = ["quantiles", "mean"]
                if intervals is None:
                    # This corresponds to quantiles: 0.05 to 0.95 in steps of
                    # 0.05.
                    params["intervals"] = list(range(0, 100, 10))
                else:
                    params["intervals"] = np.sort(intervals).tolist()

            forecast_dict, console_output = self._run_r_forecast(
                data, params, save_info=save_info
            )

            if self.method_name in QUANTILE_FORECAST_METHODS:
                quantile_forecasts_dict = forecast_dict["quantiles"]

                yield QuantileForecast(
                    forecast_arrays=np.array(
                        list(quantile_forecasts_dict.values())
                    ),
                    forecast_keys=list(quantile_forecasts_dict.keys()),
                    start_date=forecast_start(data),
                    item_id=data.get("item_id", None),
                )
            else:
                if self.method_name in POINT_FORECAST_METHODS:
                    # Handling special cases outside of R is better, since it
                    # is more visible and is easier to change. Repeat mean
                    # forecasts `num_samples` times.
                    samples = np.reshape(
                        forecast_dict["mean"] * params["num_samples"],
                        (params["num_samples"], self.prediction_length),
                    )
                else:
                    samples = np.array(forecast_dict["samples"])

                expected_shape = (
                    params["num_samples"],
                    self.prediction_length,
                )
                assert (
                    samples.shape == expected_shape
                ), f"Expected shape {expected_shape} but found {samples.shape}"
                info = (
                    {"console_output": "\n".join(console_output)}
                    if save_info
                    else None
                )
                yield SampleForecast(
                    samples,
                    forecast_start(data),
                    info=info,
                    item_id=data.get("item_id", None),
                )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import os
import tarfile
from pathlib import Path
from typing import Optional, Set


def compress_directory(
    directory: Path, target: Path, include: Optional[Set[str]] = None
) -> None:
    """
    Compresses the provided directory into a single `.tar.gz` file.

    Args:
        directory: The directory to compress.
        target: The `.tar.gz` file where the compressed archive should be written.
        include: The filenames to include. If not provided, all files are included.
    """
    with target.open("wb+") as f:
        with tarfile.open(fileobj=f, mode="w:gz") as tar:
            for root, _, files in os.walk(directory):
                for file in files:
                    if include is not None and file not in include:
                        continue
                    name = os.path.join(root, file)
                    tar.add(name, arcname=os.path.relpath(name, directory))



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import Literal, Optional
import torch
from torch import nn


class ListMLELoss(nn.Module):
    """
    Loss that is used for ListMLE.

    For each feature, ranking is performed independently. A lower score
    indicates a lower rank (i.e. "better" value).
    """

    def __init__(
        self,
        discount: Optional[
            Literal["logarithmic", "linear", "quadratic"]
        ] = None,
    ):
        """
        Args:
            discount: The type of discount to apply. If discounting, higher-ranked values are more
                important.
        """
        super().__init__()
        self.discount = discount

    def forward(
        self,
        y_pred: torch.Tensor,
        y_true: torch.Tensor,
        group_ids: torch.Tensor,
    ) -> torch.Tensor:
        """
        Computes the listwise loss of the predictions.

        Args:
            y_pred: Tensor of shape [N, D] with the predicted values (N: number of items, D:
                number of metrics).
            y_true: Tensor of shape [N, D] with the actual values.
            group_ids: Tensor of shape [N] with the group IDs for each item. All items within a
                group are ranked.

        Returns:
            Tensor of shape [1] containing the loss.
        """
        # We compute the loss for each group and subsequently average
        log_likelihoods = []
        for group_id in group_ids.unique():
            # First, we extract the values belonging to the group
            mask = group_ids == group_id
            group_pred = -y_pred[
                mask
            ]  # for ListMLE, a higher score is "better, so we invert here
            group_true = y_true[mask]

            # Then, compute numerator and denominator
            order = group_true.argsort(0)
            num = group_pred.gather(0, order)
            denom = num.flip(0).logcumsumexp(0).flip(0)

            # And compute the log likelihood
            log_likeli = num - denom

            # Optionally, we also apply the discount factor
            n = log_likeli.size(0)
            if self.discount is not None:
                if self.discount == "logarithmic":
                    denom = (torch.arange(n) + 2)[:, None].log()
                elif self.discount == "linear":
                    denom = (torch.arange(n) + 1)[:, None]
                else:  # self.discount == "quadratic"
                    denom = (torch.arange(n) + 1)[:, None] ** 2
                # We additionally scale by n / denom.sum() to scale the loss
                log_likeli = (log_likeli / denom) * (
                    n / denom.reciprocal().sum()
                )

            log_likelihoods.append(log_likeli.mean())

        # Eventually compute the loss as the NLL
        loss = -torch.stack(log_likelihoods).mean()
        return loss



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

"""
This example shows how to do anomaly detection with DeepAR.
The model is first trained and then time-points with the largest negative log-likelihood are plotted.
"""
import numpy as np
from itertools import islice
from functools import partial
import mxnet as mx
import matplotlib.pyplot as plt
import pandas as pd
from pandas.plotting import register_matplotlib_converters

register_matplotlib_converters()

from gluonts.dataset.loader import TrainDataLoader
from gluonts.model.deepar import DeepAREstimator
from gluonts.mx.util import get_hybrid_forward_input_names
from gluonts.mx.trainer import Trainer
from gluonts.mx.batchify import batchify
from gluonts.dataset.repository.datasets import get_dataset


if __name__ == "__main__":

    dataset = get_dataset(dataset_name="electricity")

    estimator = DeepAREstimator(
        prediction_length=dataset.metadata.prediction_length,
        freq=dataset.metadata.freq,
        trainer=Trainer(
            learning_rate=1e-3, epochs=20, num_batches_per_epoch=100
        ),
    )

    # instead of calling `train` method, we call `train_model` that returns more things including the training model
    train_output = estimator.train_model(dataset.train)

    input_names = get_hybrid_forward_input_names(
        type(train_output.trained_net)
    )

    # we construct a data_entry that contains 500 random windows
    batch_size = 500
    num_samples = 100
    instance_splitter = estimator._create_instance_splitter("training")
    training_data_loader = TrainDataLoader(
        dataset=dataset.train,
        transform=train_output.transformation + instance_splitter,
        batch_size=batch_size,
        num_batches_per_epoch=estimator.trainer.num_batches_per_epoch,
        stack_fn=partial(
            batchify, ctx=estimator.trainer.ctx, dtype=estimator.dtype
        ),
    )

    data_entry = next(iter(training_data_loader))

    # we now call the train model to get the predicted distribution on each window
    # this allows us to investigate where are the biggest anomalies
    context_length = train_output.trained_net.context_length
    prediction_length = train_output.trained_net.prediction_length

    distr = train_output.trained_net.distribution(
        *[data_entry[k] for k in input_names]
    )

    # gets all information into numpy array for further plotting
    samples = distr.sample(num_samples).asnumpy()
    percentiles = np.percentile(samples, axis=0, q=[10.0, 90.0])
    target = mx.ndarray.concat(
        data_entry["past_target"], data_entry["future_target"], dim=1
    )
    target = target[:, -(context_length + prediction_length) :]
    nll = -distr.log_prob(target).asnumpy()
    target = target.asnumpy()
    mean = samples.mean(axis=0)
    percentiles = np.percentile(samples, axis=0, q=[10.0, 90.0])

    # NLL indices from largest to smallest
    sorted_indices = np.argsort(nll.reshape(-1))[::-1]

    # shows the series and times when the 20 largest NLL were observed
    for k in sorted_indices[:20]:
        i = k // nll.shape[1]
        t = k % nll.shape[1]

        time_index = pd.period_range(
            data_entry["forecast_start"][i],
            periods=context_length + prediction_length,
            freq=data_entry["forecast_start"][i].freq,
        )
        time_index -= context_length * time_index.freq

        plt.figure(figsize=(10, 4))
        plt.fill_between(
            time_index,
            percentiles[0, i],
            percentiles[-1, i],
            alpha=0.5,
            label="80% CI predicted",
        )
        plt.plot(time_index, target[i], label="target")
        plt.axvline(time_index[t], alpha=0.5, color="r")
        plt.title(f"NLL: {nll[i, t]}")
        plt.legend()
        plt.show()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import itertools
import math
import random
from pathlib import Path
from typing import Callable, Dict, List, Optional, Tuple, Type
import numpy as np
import pandas as pd
from tqdm.auto import tqdm
from tsbench.analysis.utils import (
    loocv_split,
    num_fitting_processes,
    run_parallel,
)
from tsbench.config import (
    Config,
    DATASET_REGISTRY,
    DatasetConfig,
    ModelConfig,
    TrainConfig,
)
from tsbench.config.model.models import SeasonalNaiveModelConfig
from tsbench.constants import DEFAULT_DATA_PATH
from tsbench.evaluations.metrics.performance import Metric, Performance
from tsbench.evaluations.tracking import ModelTracker
from tsbench.forecasts import (
    ensemble_forecasts,
    EnsembleWeighting,
    evaluate_forecasts,
    Evaluation,
)


class EnsembleAnalyzer:
    """
    The ensemble analyzer allows for evaluating the performance of ensembles
    across datasets.

    The analysis is run in parallel and should, thus, not be used in a Jupyter
    notebook. Instead, consider using the `tsbench` CLI.
    """

    def __init__(
        self,
        tracker: ModelTracker,
        ensemble_size: Optional[int] = 10,
        ensemble_weighting: EnsembleWeighting = "uniform",
        config_class: Optional[Type[ModelConfig]] = None,
    ):
        """
        Args:
            tracker: The tracker from which to obtain pretrained models and forecasts.
            ensemble_size: The number of models to use when building an ensemble. If not provided,
                uses as many models as possible.
            ensemble_weighting: The type of ensemble weighting to use for averaging forecasts.
            config_class: The class of models to ensemble. If this is provided, fewer models than
                the given ensemble size might be selected.
        """
        self.tracker = tracker
        self.ensemble_size = ensemble_size
        self.ensemble_weighting: EnsembleWeighting = ensemble_weighting
        self.config_class = config_class

    def run(self) -> Tuple[pd.DataFrame, Dict[str, List[ModelConfig]]]:
        """
        Runs the evaluation on the data provided via the tracker. The data
        obtained from the tracker is partitioned by the dataset and we run
        "grouped LOOCV" to compute performance metrics on datasets. Metrics on
        each dataset are then returned as data frame.

        Returns:
            The metrics on the individual datasets.
            The model choices for each dataset.
        """
        results = run_parallel(
            self._run_on_dataset,
            data=list(loocv_split(self.tracker)),
            num_processes=num_fitting_processes(),
        )
        performances = [r[0] for r in results]
        member_mapping = {k: v for r in results for k, v in r[1].items()}

        df = pd.concat(performances).set_index("test_dataset")
        return df, member_mapping

    def _run_on_dataset(
        self,
        data: Tuple[
            Tuple[List[Config[ModelConfig]], List[Performance]],
            Tuple[List[Config[ModelConfig]], List[Performance]],
        ],
    ) -> Tuple[pd.DataFrame, Dict[str, List[ModelConfig]]]:
        # Extract the data
        _, (X_test, y_test) = data

        # Compute the metrics
        performance, members = self._performance_on_dataset(X_test, y_test)

        # Transform into output
        df = Performance.to_dataframe([performance]).assign(
            test_dataset=X_test[0].dataset.name()
        )
        return df, {X_test[0].dataset.name(): members}

    def _performance_on_dataset(
        self,
        X_test: List[Config[ModelConfig]],
        y_test: List[Performance],
    ) -> Tuple[Performance, List[ModelConfig]]:
        # If there are model constraints, we restrict the test set
        if self.config_class is not None:
            indices = [
                i
                for i, c in enumerate(X_test)
                if isinstance(c.model, self.config_class)
                and (
                    not isinstance(c.model, TrainConfig)
                    or c.model.training_fraction == 1
                )
            ]
            X_test = [X_test[i] for i in indices]
            y_test = [y_test[i] for i in indices]

        # After that, we sort by the validation or test nCRPS and pick the ensemble members
        val_scores = [self.tracker.get_validation_scores(x) for x in X_test]
        if any(x is None for x in val_scores):
            order = np.argsort([p.ncrps.mean for p in y_test]).tolist()
        else:
            order = np.argsort([s.ncrps for s in val_scores])  # type: ignore
        choices = order[: (self.ensemble_size or len(order))]

        # In case no model is selected, seasonal naïve is selected (this should always be
        # fast enough to satisfy the latency constraint unless our measurement is faulty or
        # the implementation inefficient)
        if len(choices) == 0:
            choices = [
                i
                for i, c in enumerate(X_test)
                if isinstance(c.model, SeasonalNaiveModelConfig)
            ]

        # Eventually, we get the ensemble members and compute the ensemble performance
        members = [X_test[i].model for i in choices]
        performance = self.get_ensemble_performance(
            members,
            dataset=X_test[0].dataset,
            member_performances=[y_test[i] for i in choices],
        )

        return performance, members

    # ---------------------------------------------------------------------------------------------

    def get_ensemble_performance(
        self,
        models: List[ModelConfig],
        dataset: DatasetConfig,
        member_performances: Optional[List[Performance]] = None,
        num_samples: int = 10,
    ) -> Performance:
        """
        Estimates the performance of a list of models on a particular dataset.
        For this, actually trained models are sampled for each configuration.

        Args:
            models: The list of models to evaluate.
            dataset: The dataset to evaluate on.
            member_performances: The (predicted) performances of the provided models. Used to weigh
                the ensemble members. If not provided, uses the true performances.
            num_samples: The number of samples for estimating the performance.

        Returns:
            The expected performance of the ensemble.
        """
        if member_performances is None:
            member_performances = [
                self.tracker.get_performance(Config(m, dataset))
                for m in models
            ]

        # First, we need to get the forecasts for all models
        forecasts = [
            self.tracker.get_forecasts(Config(m, dataset)) for m in models
        ]

        # Then, we want to construct min(#available_choices, 10) different ensembles by randomly
        # choosing models from the configurations without replacement.
        max_choices = np.prod([len(f) for f in forecasts])
        num_choices = min(max_choices, num_samples)
        pool = itertools.product(*[range(len(f)) for f in forecasts])
        model_combinations = random.sample(list(pool), k=num_choices)

        # Then, we evaluate each of the ensembles
        evaluations = []
        for combination in model_combinations:
            ensembled_forecast = ensemble_forecasts(
                [f[i] for i, f in zip(combination, forecasts)],
                self.ensemble_weighting,
                [p.ncrps.mean for p in member_performances],
            )
            evaluation = evaluate_forecasts(
                ensembled_forecast, dataset.data.test().evaluation()
            )
            evaluations.append(evaluation)

        # And eventually, we build the resulting performance object
        performance = Evaluation.performance(evaluations)
        performance.num_gradient_updates = self._combine_metrics(
            member_performances, lambda p: p.num_gradient_updates
        )
        performance.num_model_parameters = self._combine_metrics(
            member_performances, lambda p: p.num_model_parameters
        )
        performance.latency = self._combine_metrics(
            member_performances, lambda p: p.latency
        )
        performance.training_time = self._combine_metrics(
            member_performances, lambda p: p.training_time
        )
        return performance

    def _combine_metrics(
        self,
        performances: List[Performance],
        metric: Callable[[Performance], Metric],
    ) -> Metric:
        return Metric(
            mean=sum(metric(p).mean for p in performances),
            std=math.sqrt(sum(metric(p).std ** 2 for p in performances)),
        )

    # ---------------------------------------------------------------------------------------------

    def get_all_performances(
        self,
        models: Dict[str, List[ModelConfig]],
        data_path: Path = DEFAULT_DATA_PATH,
    ) -> pd.DataFrame:
        """
        TODO.
        """
        results = []
        for name, dataset in tqdm(DATASET_REGISTRY.items()):
            performance = self.get_ensemble_performance(
                models[name], dataset(data_path)
            )
            df = Performance.to_dataframe([performance]).assign(
                test_dataset=name
            )
            results.append(df)

        return pd.concat(results).set_index("test_dataset")



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import os
import pickle
from pathlib import Path
from typing import Dict, List
from tsbench.config import Config, EnsembleConfig
from tsbench.evaluations.metrics import Performance
from ._base import Tracker
from ._evaluations import Evaluations


class EnsembleTracker(Tracker[EnsembleConfig]):
    """
    Tracker which sources information from .pickle files providing
    configurations of ensemble members and the ensembles' performances.
    """

    def __init__(self, directory: Path):
        """
        Args:
            files: Directory from which to load the files non-recursively.
        """
        configurations = []
        performances = []
        for file in os.listdir(directory):
            if not file.endswith(".pickle"):
                continue
            with Path(file).open("rb") as f:
                data = pickle.load(f)
                configurations.extend(
                    [
                        Config(frozenset(x["configurations"]), x["dataset"])
                        for x in data
                    ]
                )
                performances.extend([x["performance"] for x in data])

        self.performance_map: Dict[Config[EnsembleConfig], Performance] = dict(
            zip(configurations, performances)
        )

    def unique_ensembles(self) -> List[EnsembleConfig]:
        """
        Returns the unique configurations of ensembles provided by this
        tracker.
        """
        return list({c.model for c in self.performance_map})

    def get_evaluations(self) -> Evaluations[EnsembleConfig]:
        return Evaluations(
            list(self.performance_map),
            list(self.performance_map.values()),
        )

    def get_performance(self, config: Config[EnsembleConfig]) -> Performance:
        return self.performance_map[config]



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List, Literal, Optional
import numpy as np
import numpy.typing as npt
from sklearn.multioutput import MultiOutputRegressor
from sklearn.preprocessing import LabelEncoder
from xgboost.sklearn import XGBRanker, XGBRegressor
from tsbench.config import Config, ModelConfig
from tsbench.evaluations.tracking import ModelTracker
from ._base import DatasetFeaturesMixin, OutputNormalization, Surrogate
from ._factory import register_surrogate
from .transformers import ConfigTransformer


@register_surrogate("xgboost")
class XGBoostSurrogate(Surrogate[ModelConfig], DatasetFeaturesMixin):
    """
    The XGBoost surrogate predicts a model's performance on a new dataset by
    using independent XGBoost regressors for each performance metric.

    For this, models and hyperparameters are converted to feature vectors.
    """

    def __init__(
        self,
        tracker: ModelTracker,
        objective: Literal["regression", "ranking"] = "regression",
        use_simple_dataset_features: bool = False,
        use_seasonal_naive_performance: bool = False,
        use_catch22_features: bool = False,
        predict: Optional[List[str]] = None,
        output_normalization: OutputNormalization = None,
        impute_simulatable: bool = False,
    ):
        """
        Args:
            tracker: A tracker that can be used to impute latency and number of model parameters
                into model performances. Also, it is required for some input features.
            objective: The optimization objective for the XGBoost estimators.
            use_simple_dataset_features: Whether to use dataset features to predict using a
                weighted average.
            use_seasonal_naive_performance: Whether to use the Seasonal Naïve nCRPS as dataset
                featuers. Requires the cacher to be set.
            use_catch22_features: Whether to use catch22 features for datasets statistics. Ignored
                if `use_dataset_features` is not set.
            predict: The metrics to predict. All if not provided.
            output_normalization: The type of normalization to apply to the features of each
                dataset independently. `None` applies no normalization, "quantile" applies quantile
                normalization, and "standard" transforms data to have zero mean and unit variance.
            impute_simulatable: Whether the tracker should impute latency and number of model
                parameters into the returned performance object.
        """
        super().__init__(
            tracker, predict, output_normalization, impute_simulatable
        )

        self.use_ranking = objective == "ranking"
        self.config_transformer = ConfigTransformer(
            add_model_features=True,
            add_dataset_statistics=use_simple_dataset_features,
            add_seasonal_naive_performance=use_seasonal_naive_performance,
            add_catch22_features=use_catch22_features,
            tracker=tracker,
        )

        if self.use_ranking:
            base_estimator = XGBRanker(objective="rank:pairwise", nthread=4)
        else:
            base_estimator = XGBRegressor(nthread=4)
        self.estimator = MultiOutputRegressor(base_estimator)

    @property
    def required_cpus(self) -> int:
        return 4

    def _fit(
        self, X: List[Config[ModelConfig]], y: npt.NDArray[np.float32]
    ) -> None:
        X_numpy = self.config_transformer.fit_transform(X)

        if self.use_ranking:
            # We need to sort by dataset
            encoder = LabelEncoder()
            dataset_indices = encoder.fit_transform(
                [x.dataset.name() for x in X]
            )
            sorting = np.argsort(dataset_indices)

            # Then, sort X and y and assign the group IDs
            X_grouped = [X_numpy[i] for i in sorting]
            y_grouped = [y[i] for i in sorting]
            self.estimator.fit(
                X_grouped, y_grouped, qid=dataset_indices[sorting]
            )
        else:
            self.estimator.fit(X_numpy, y)

    def _predict(
        self, X: List[Config[ModelConfig]]
    ) -> npt.NDArray[np.float32]:
        X_numpy = self.config_transformer.transform(X)
        return self.estimator.predict(X_numpy)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

"""
This example shows how to fit a model and evaluate its predictions.
"""

# third party imports
import pandas as pd
import pytest

# first party imports
from gluonts.dataset.artificial import constant_dataset
from gluonts.dataset.common import ListDataset
from gluonts.dataset.field_names import FieldName
from gluonts.dataset.rolling_dataset import (
    StepStrategy,
    generate_rolling_dataset,
)


def generate_dataset(name):
    dataset = None
    if name == "constant":
        _, _, dataset = constant_dataset()
    elif name == "varying":
        # Tests edge cases
        # t0: start time of target
        # ts: start time of desired range
        # te: end time of desired range
        # t1: end time of target
        # ts < te, t0 <= t1
        #
        # start time index of rolling window is 20
        # end time index of rolling window is 24
        # ts = 2000-01-01 20:00:00
        # te = 2000-01-02 00:00:00

        ds_list = [
            {  # test 1: ends after end time, te > t1
                "target": [0.0] * 30,
                "start": pd.Timestamp(2000, 1, 1, 0, 0),
            },
            {  # test 2: ends at the end time, te == t1
                "target": [0.0] * 25,
                "start": pd.Timestamp(2000, 1, 1, 0, 0),
            },
            {  # test 3: ends between start and end times, ts < t1 < te
                "target": [0.0] * 23,
                "start": pd.Timestamp(2000, 1, 1, 0, 0),
            },
            {  # test 4: ends on start time, ts == t1
                "target": [0.0] * 20,
                "start": pd.Timestamp(2000, 1, 1, 0, 0),
            },
            {  # test 5: ends before start time, t1 < ts
                "target": [0.0] * 15,
                "start": pd.Timestamp(2000, 1, 1, 0, 0),
            },
            {  # test 6: starts on start ends after end, ts == t0, te > t1
                "target": [0.0] * 10,
                "start": pd.Timestamp(2000, 1, 1, 20, 0),
            },
            {  # test 7: starts in between ts and te, ts < t0 < te < t1
                "target": [0.0] * 10,
                "start": pd.Timestamp(2000, 1, 1, 22, 0),
            },
            {  # test 8: starts on end time, te == t0
                "target": [0.0] * 10,
                "start": pd.Timestamp(2000, 1, 2, 0, 0),
            },
            {  # test 9: starts after end time, te < t0
                "target": [0.0] * 10,
                "start": pd.Timestamp(2000, 1, 2, 1, 0),
            },
            {  # test 10: starts after ts & ends before te, ts < t0 < t1 < te
                "target": [0.0] * 3,
                "start": pd.Timestamp(2000, 1, 1, 21, 0),
            },
        ]
        dataset = ListDataset(ds_list, "H")
    else:
        pytest.raises(ValueError)
    return dataset


@pytest.mark.parametrize(
    "prediction_length, unique",
    [(p, u) for p in [-1, 0] for u in [True, False]],
)
def test_invalid_rolling_parameters(prediction_length, unique):
    try:
        generate_rolling_dataset(
            dataset=generate_dataset("constant"),
            start_time=pd.Period("2000-01-01-20", freq="1H"),
            end_time=pd.Period("2000-01-02-00", freq="1H"),
            strategy=StepStrategy(
                step_size=prediction_length if unique else 1,
                prediction_length=prediction_length,
            ),
        )
        # program should have failed at this point
        pytest.raises(RuntimeError)
    except AssertionError:
        pass


@pytest.mark.parametrize(
    "ds_name, prediction_length, unique, ignore_end, ds_expected",
    [
        (
            "varying",
            2,
            False,
            False,
            [[0.0] * length for length in range(25, 21, -1)] * 2
            + [[0.0] * length for length in [23, 22]]
            + [[0.0] * length for length in range(5, 1, -1)]
            + [[0.0] * length for length in [3, 2]] * 2,
        ),
        (
            "varying",
            2,
            True,
            False,
            [
                [0.0] * 25,
                [0.0] * 23,
                [0.0] * 25,
                [0.0] * 23,
                [0.0] * 23,
                [0.0] * 5,
                [0.0] * 3,
                [0.0] * 3,
                [0.0] * 3,
            ],
        ),
        (
            "varying",
            3,
            False,
            True,
            [[0.0] * length for length in range(30, 22, -1)]
            + [[0.0] * length for length in [25, 24, 23]]
            + [[0.0] * 23]
            + [[0.0] * length for length in range(10, 2, -1)] * 4
            + [[0.0] * 3],
        ),
        (
            "constant",
            1,
            True,
            False,
            [
                [float(val)] * length
                for val in range(10)
                for length in [25, 24, 23, 22, 21]
            ],
        ),
        (
            "constant",
            1,
            False,
            False,
            [
                [float(val)] * length
                for val in range(10)
                for length in [25, 24, 23, 22, 21]
            ],
        ),
        (
            "constant",
            2,
            True,
            False,
            [
                [float(val)] * length
                for val in range(10)
                for length in [25, 23]
            ],
        ),
        (
            "constant",
            2,
            False,
            False,
            [
                [float(val)] * length
                for val in range(10)
                for length in [25, 24, 23, 22]
            ],
        ),
        (
            "constant",
            3,
            True,
            False,
            [[float(val)] * 25 for val in range(10)],
        ),
        (
            "constant",
            3,
            False,
            False,
            [
                [float(val)] * length
                for val in range(10)
                for length in [25, 24, 23]
            ],
        ),
        (
            "constant",
            4,
            True,
            False,
            [[float(val)] * 25 for val in range(10)],
        ),
        (
            "constant",
            4,
            False,
            False,
            [
                [float(val)] * length
                for val in range(10)
                for length in [25, 24]
            ],
        ),
    ],
)
def test_step_strategy(
    ds_name, prediction_length, unique, ignore_end, ds_expected
):
    rolled_ds = generate_rolling_dataset(
        dataset=generate_dataset(ds_name),
        start_time=pd.Period("2000-01-01-20", freq="1H"),
        end_time=None if ignore_end else pd.Period("2000-01-02-00", freq="1H"),
        strategy=StepStrategy(
            step_size=prediction_length if unique else 1,
            prediction_length=prediction_length,
        ),
    )

    i = 0
    for ts in rolled_ds:
        assert len(ts["target"]) == len(ds_expected[i])
        for rolled_result, expected in zip(ts["target"], ds_expected[i]):
            assert rolled_result == expected
        i += 1

    assert len(ds_expected) == i


def create_dynamic_dataset(
    start: str, length: int, num_dynamic: int
) -> ListDataset:
    """Create a ListDataset with dynamic values equal to the target."""
    return ListDataset(
        [
            {
                FieldName.TARGET: list(range(length)),
                FieldName.START: pd.Timestamp(start),
                FieldName.FEAT_DYNAMIC_REAL: [list(range(length))]
                * num_dynamic,
                FieldName.FEAT_DYNAMIC_CAT: [list(range(length))]
                * num_dynamic,
            }
        ],
        freq="D",
    )


@pytest.mark.parametrize(
    "test_length, prediction_length, target_start, rolling_start, num_dynamic_feat",
    [
        (15, 2, "01-01-2019", "01-13-2019", 1),
        (15, 2, "01-01-2019", "01-13-2019", 2),
        (15, 1, "01-01-2019", "01-10-2019", 1),
        (15, 1, "01-01-2019", "01-10-2019", 2),
    ],
)
def test_dynamic_features(
    test_length: int,
    prediction_length: int,
    target_start: str,
    rolling_start: str,
    num_dynamic_feat: int,
):
    # generate datasets with dynamic values equal to the target values
    rolled_ds = generate_rolling_dataset(
        dataset=create_dynamic_dataset(
            target_start, test_length, num_dynamic_feat
        ),
        strategy=StepStrategy(prediction_length=prediction_length),
        start_time=pd.Period(rolling_start),
    )

    for timeseries in rolled_ds:
        for feature_name in (
            FieldName.FEAT_DYNAMIC_REAL,
            FieldName.FEAT_DYNAMIC_CAT,
        ):
            assert len(timeseries[feature_name]) == num_dynamic_feat
            for feature in timeseries[feature_name]:
                # since target and dynamic features has the same values
                # in the dataset we can ensure ordering and length of
                # the features by comparing them with the target value
                assert (timeseries[FieldName.TARGET] == feature).all()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from pathlib import Path

# Default GluonTS version for when the framework version is not specified.
# This is no longer updated so as to not break existing workflows.
GLUONTS_VERSION = "0.4.1"

# Framework related
FRAMEWORK_NAME = "GluonTS"
LOWEST_MMS_VERSION = "1.4"
LOWEST_SCRIPT_MODE_VERSION = "0", "4", "1"
LATEST_GLUONTS_VERSION = "0.4.1"
PYTHON_VERSION = "py3"

# Training related
ENTRY_POINTS_FOLDER = Path(__file__).parent.resolve() / "entry_point_scripts"
TRAIN_SCRIPT = "train_entry_point.py"
MONITORED_METRICS = "mean_wQuantileLoss", "ND", "RMSE"
NUM_SAMPLES = 100
QUANTILES = 0.1, 0.2, 0.3, 0.4, 0.5, 0.6, 0.7, 0.8, 0.9



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

"""
This example shows how to serialize and deserialize a model
"""
import os
import pprint

from gluonts.dataset.repository.datasets import get_dataset, get_download_path
from gluonts.evaluation import make_evaluation_predictions, Evaluator
from gluonts.model.simple_feedforward import SimpleFeedForwardEstimator
from gluonts.mx.trainer import Trainer
from gluonts.model.predictor import Predictor

if __name__ == "__main__":

    dataset = get_dataset("exchange_rate")

    estimator = SimpleFeedForwardEstimator(
        prediction_length=dataset.metadata.prediction_length,
        trainer=Trainer(epochs=5, num_batches_per_epoch=10),
    )

    predictor = estimator.train(dataset.train)

    # save the trained model in a path ~/.mxnet/gluon-ts/feedforward/
    # or $MXNET_HOME/feedforward if MXNET_HOME is defined
    model_path = get_download_path() / "feedforward"
    os.makedirs(model_path, exist_ok=True)

    predictor.serialize(model_path)

    # loads it back and evaluate predictions accuracy with the deserialized model
    predictor_deserialized = Predictor.deserialize(model_path)

    forecast_it, ts_it = make_evaluation_predictions(
        dataset.test, predictor=predictor_deserialized, num_samples=100
    )

    agg_metrics, item_metrics = Evaluator()(
        ts_it, forecast_it, num_series=len(dataset.test)
    )

    pprint.pprint(agg_metrics)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import pickle
from io import BytesIO
from typing import Any, Dict, List
import pandas as pd
from gridfs import GridFS
from pymongo.database import Database


class SacredExperiment:
    """
    A sacred experiment describes a Sacred experiment stored in MongoDB and is
    retrieved from a Sacred Mongo client.
    """

    def __init__(self, info: Dict[str, Any], db: Database, gridfs: GridFS):
        """
        **Not to be used manually.**
        """
        self.info = info
        self.db = db
        self.fs = gridfs

    @property
    def config(self) -> Dict[str, Any]:
        """
        Returns the configuration of the experiment.
        """
        return self.info["config"]

    @property
    def artifacts(self) -> List[str]:
        """
        Returns the names of all artifacts associated with the experiment.
        """
        return [a["name"] for a in self.info["artifacts"]]

    def read_parquet(self, artifact: str) -> pd.DataFrame:
        """
        Reads the parquet file from the artifact with the specified name.

        Args:
            artifact: The name of the artifact.

        Returns:
            The parquet file loaded as data frame.
        """
        matches = [
            a["file_id"]
            for a in self.info["artifacts"]
            if a["name"] == artifact
        ]
        data = self.fs.get(matches[0]).read()
        return pd.read_parquet(BytesIO(data))

    def read_pickle(self, artifact: str) -> Any:
        """
        Reads the pickled file from the artifact with the specified name.

        Args:
            artifact: The name of the artifact.

        Returns:
            The data that was pickled.
        """
        matches = [
            a["file_id"]
            for a in self.info["artifacts"]
            if a["name"] == artifact
        ]
        data = self.fs.get(matches[0]).read()
        return pickle.loads(data)

    def delete(self) -> None:
        """
        Deletes the experiment by setting the associated experiment name to
        "Trash".
        """
        self.db.runs.update_one(
            {"_id": self.info["_id"]},
            {"$set": {"config.name": "Trash"}},
            upsert=False,
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(config={self.info['config']})"



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from copy import deepcopy
from dataclasses import dataclass, field
from typing import Any, cast, Dict, Iterator, List, Optional, Union

import pandas as pd
from pandas.core.indexes.datetimelike import DatetimeIndexOpsMixin
from toolz import valmap

from gluonts.dataset.common import DataEntry, ProcessDataEntry
from gluonts.dataset.field_names import FieldName


@dataclass
class PandasDataset:
    """
    A pandas.DataFrame-based dataset type.

    This class is constructed with a collection of pandas.DataFrame-objects
    where each ``DataFrame`` is representing one time series.
    A ``target`` and a ``timestamp`` columns are essential. Furthermore,
    static/dynamic real/categorical features can be specified.

    Parameters
    ----------
    dataframes
        Single ``pd.DataFrame``/``pd.Series`` or a collection as list or dict
        containing at least ``timestamp`` and ``target`` values.
        If a Dict is provided, the key will be the associated ``item_id``.
    target
        Name of the column that contains the ``target`` time series.
        For multivariate targets, a list of column names should be provided.
    timestamp
        Name of the column that contains the timestamp information.
    freq
        Frequency of observations in the time series. Must be a valid pandas
        frequency.
    feat_dynamic_real
        List of column names that contain dynamic real features.
    feat_dynamic_cat
        List of column names that contain dynamic categorical features.
    feat_static_real
        List of column names that contain static real features.
    feat_static_cat
        List of column names that contain static categorical features.
    past_feat_dynamic_real
        List of column names that contain dynamic real features only for the
        history.
    ignore_last_n_targets
        For target and past dynamic features last ``ignore_last_n_targets``
        elements are removed when iterating over the data set. This becomes
        important when the predictor is called.
    """

    dataframes: Union[
        pd.DataFrame,
        pd.Series,
        List[pd.DataFrame],
        List[pd.Series],
        Dict[str, pd.DataFrame],
        Dict[str, pd.Series],
    ]
    target: Union[str, List[str]] = "target"
    timestamp: Optional[str] = None
    freq: Optional[str] = None
    feat_dynamic_real: List[str] = field(default_factory=list)
    feat_dynamic_cat: List[str] = field(default_factory=list)
    feat_static_real: List[str] = field(default_factory=list)
    feat_static_cat: List[str] = field(default_factory=list)
    past_feat_dynamic_real: List[str] = field(default_factory=list)
    ignore_last_n_targets: int = 0

    def __post_init__(self) -> None:
        if isinstance(self.target, list) and len(self.target) == 1:
            self.target = self.target[0]
        self.one_dim_target = not isinstance(self.target, list)

        if is_series(self.dataframes):
            self.dataframes = series_to_dataframe(self.dataframes)
        # store data internally as List[Tuple[str, pandas.DataFrame]]
        # if str is not empty it will be set in ``DataEntry`` as ``item_id``.
        if isinstance(self.dataframes, dict):
            self._dataframes = list(self.dataframes.items())
        elif isinstance(self.dataframes, list):
            self._dataframes = [(None, df) for df in self.dataframes]
        else:  # case single dataframe
            self._dataframes = [(None, self.dataframes)]

        for i, (item_id, df) in enumerate(self._dataframes):
            if self.timestamp:
                df = df.set_index(keys=self.timestamp)

            if not isinstance(df.index, pd.PeriodIndex):
                df.index = pd.to_datetime(df.index)
                df = df.to_period(freq=self.freq)

            df.sort_index(inplace=True)

            assert is_uniform(df.index), (
                "Dataframe index is not uniformly spaced. "
                "If your dataframe contains data from multiple series in the "
                'same column ("long" format), consider constructing the '
                "dataset with `PandasDataset.from_long_dataframe` instead."
            )

            self._dataframes[i] = (item_id, df)

        if not self.freq:  # infer frequency from index
            self.freq = self._dataframes[0][1].index.freqstr

        self.process = ProcessDataEntry(
            cast(str, self.freq), one_dim_target=self.one_dim_target
        )

    def _dataentry(
        self, item_id: Optional[str], df: pd.DataFrame
    ) -> DataEntry:
        dataentry = as_dataentry(
            data=df,
            target=self.target,
            feat_dynamic_real=self.feat_dynamic_real,
            feat_dynamic_cat=self.feat_dynamic_cat,
            feat_static_real=self.feat_static_real,
            feat_static_cat=self.feat_static_cat,
            past_feat_dynamic_real=self.past_feat_dynamic_real,
        )
        if item_id is not None:
            dataentry["item_id"] = item_id
        return dataentry

    def __iter__(self) -> Iterator[DataEntry]:
        for item_id, df in self._dataframes:
            dataentry = self.process(self._dataentry(item_id, df))
            if self.ignore_last_n_targets:
                dataentry = prepare_prediction_data(
                    dataentry, self.ignore_last_n_targets
                )
            yield dataentry

    def __len__(self) -> int:
        return len(self._dataframes)

    @classmethod
    def from_long_dataframe(
        cls, dataframe: pd.DataFrame, item_id: str, **kwargs
    ) -> "PandasDataset":
        """
        Construct ``PandasDataset`` out of a long dataframe.
        A long dataframe uses the long format for each variable. Target time
        series values, for example, are stacked on top of each other rather
        than side-by-side. The same is true for other dynamic or categorical
        features.

        Parameters
        ----------
        dataframe
            pandas.DataFrame containing at least ``timestamp``, ``target`` and
            ``item_id`` columns.
        item_id
            Name of the column that, when grouped by, gives the different time
            series.
        **kwargs
            Additional arguments. Same as of PandasDataset class.

        Returns
        -------
        PandasDataset
            Gluonts dataset based on ``pandas.DataFrame``s.
        """
        return cls(dataframes=dict(list(dataframe.groupby(item_id))), **kwargs)


def series_to_dataframe(
    series: Union[pd.Series, List[pd.Series], Dict[str, pd.Series]]
) -> Union[pd.DataFrame, List[pd.DataFrame], Dict[str, pd.DataFrame]]:
    def to_df(series):
        assert isinstance(
            series.index, DatetimeIndexOpsMixin
        ), "series index has to be a DatetimeIndex."
        return series.to_frame(name="target")

    if isinstance(series, list):
        return list(map(to_df, series))
    elif isinstance(series, dict):
        return valmap(to_df, series)
    return to_df(series)


def is_series(series: Any) -> bool:
    """
    return True if ``series`` is ``pd.Series`` or a collection of
    ``pd.Series``.
    """
    if isinstance(series, list):
        return is_series(series[0])
    elif isinstance(series, dict):
        return is_series(list(series.values()))
    return isinstance(series, pd.Series)


def as_dataentry(
    data: pd.DataFrame,
    target: Union[str, List[str]],
    timestamp: Optional[str] = None,
    feat_dynamic_real: List[str] = [],
    feat_dynamic_cat: List[str] = [],
    feat_static_real: List[str] = [],
    feat_static_cat: List[str] = [],
    past_feat_dynamic_real: List[str] = [],
) -> DataEntry:
    """
    Convert a single time series (uni- or multi-variate) that is given in
    a pandas.DataFrame format to a DataEntry.

    Parameters
    ----------
    data
        pandas.DataFrame containing at least ``timestamp``, ``target`` and
        ``item_id`` columns.
    target
        Name of the column that contains the ``target`` time series.
        For multivariate targets ``target`` is expecting a list of column
        names.
    timestamp
        Name of the column that contains the timestamp information.
        If ``None`` the index of ``data`` is assumed to be the time.
    feat_dynamic_real
        List of column names that contain dynamic real features.
    feat_dynamic_cat
        List of column names that contain dynamic categorical features.
    feat_static_real
        List of column names that contain static real features.
    feat_static_cat
        List of column names that contain static categorical features.
    past_feat_dynamic_real
        List of column names that contain dynamic real features only for
        the history.

    Returns
    -------
    DataEntry
        A dictionary with at least ``target`` and ``start`` field.
    """
    start = data.loc[:, timestamp].iloc[0] if timestamp else data.index[0]
    dataentry = {FieldName.START: start}

    def set_field(fieldname, col_names, f=lambda x: x):
        if col_names:
            dataentry[fieldname] = [
                f(data.loc[:, n].to_list()) for n in col_names
            ]

    if isinstance(target, str):
        dataentry[FieldName.TARGET] = data.loc[:, target].to_list()
    else:
        set_field(FieldName.TARGET, target)
    set_field(FieldName.FEAT_DYNAMIC_REAL, feat_dynamic_real)
    set_field(FieldName.FEAT_DYNAMIC_CAT, feat_dynamic_cat)
    set_field(FieldName.FEAT_STATIC_REAL, feat_static_real, lambda x: x[0])
    set_field(FieldName.FEAT_STATIC_CAT, feat_static_cat, lambda x: x[0])
    set_field(FieldName.PAST_FEAT_DYNAMIC_REAL, past_feat_dynamic_real)
    return dataentry


def prepare_prediction_data(
    dataentry: DataEntry, ignore_last_n_targets: int
) -> DataEntry:
    """
    Remove ``ignore_last_n_targets`` values from ``target`` and
    ``past_feat_dynamic_real``.  Works in univariate and multivariate case.

    >>> prepare_prediction_data(
    >>>    {"target": np.array([1., 2., 3., 4.])}, ignore_last_n_targets=2
    >>> )
    {'target': array([1., 2.])}
    """
    entry = deepcopy(dataentry)
    for fname in [FieldName.TARGET, FieldName.PAST_FEAT_DYNAMIC_REAL]:
        if fname in entry:
            entry[fname] = entry[fname][..., :-ignore_last_n_targets]
    return entry


def is_uniform(index: pd.PeriodIndex) -> bool:
    """
    Check if ``index`` contains monotonically increasing periods, evenly spaced
    with frequency ``index.freq``.

    >>> ts = ["2021-01-01 00:00", "2021-01-01 02:00", "2021-01-01 04:00"]
    >>> is_uniform(pd.DatetimeIndex(ts).to_period("2H"))
    True
    >>> ts = ["2021-01-01 00:00", "2021-01-01 04:00"]
    >>> is_uniform(pd.DatetimeIndex(ts).to_period("2H"))
    False
    """
    other = pd.period_range(index[0], periods=len(index), freq=index.freq)
    return (other == index).all()



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import functools
import logging
import shutil
from pathlib import Path
from types import ModuleType
from typing import (
    Callable,
    Iterable,
    List,
    NamedTuple,
    Optional,
    Union,
    cast,
)

import numpy as np
import pandas as pd
from pandas.tseries.frequencies import to_offset

import pydantic

from gluonts import json
from gluonts.itertools import Cached, Map
from gluonts.dataset.field_names import FieldName
from gluonts.exceptions import GluonTSDataError


from . import Dataset, DatasetCollection, DataEntry, DataBatch  # noqa
from . import jsonl, DatasetWriter


arrow: Optional[ModuleType]

try:
    from . import arrow
except ImportError:
    arrow = None


class BasicFeatureInfo(pydantic.BaseModel):
    name: str


class CategoricalFeatureInfo(pydantic.BaseModel):
    name: str
    cardinality: str


class MetaData(pydantic.BaseModel):
    freq: str = pydantic.Field(..., alias="time_granularity")  # type: ignore
    target: Optional[BasicFeatureInfo] = None

    feat_static_cat: List[CategoricalFeatureInfo] = []
    feat_static_real: List[BasicFeatureInfo] = []
    feat_dynamic_real: List[BasicFeatureInfo] = []
    feat_dynamic_cat: List[CategoricalFeatureInfo] = []

    prediction_length: Optional[int] = None

    class Config(pydantic.BaseConfig):
        allow_population_by_field_name = True


class SourceContext(NamedTuple):
    source: str
    row: int


class TrainDatasets(NamedTuple):
    """
    A dataset containing two subsets, one to be used for training purposes, and
    the other for testing purposes, as well as metadata.
    """

    metadata: MetaData
    train: Dataset
    test: Optional[Dataset] = None

    def save(
        self,
        path_str: str,
        writer: DatasetWriter,
        overwrite=False,
    ) -> None:
        """
        Saves an TrainDatasets object to a JSON Lines file.

        Parameters
        ----------
        path_str
            Where to save the dataset.
        overwrite
            Whether to delete previous version in this folder.
        """
        path = Path(path_str)

        if overwrite:
            shutil.rmtree(path, ignore_errors=True)

        path.mkdir(parents=True)
        with open(path / "metadata.json", "wb") as out_file:
            json.bdump(self.metadata.dict(), out_file, nl=True)

        train = path / "train"
        train.mkdir(parents=True)
        writer.write_to_folder(self.train, train)

        if self.test is not None:
            test = path / "test"
            test.mkdir(parents=True)
            writer.write_to_folder(self.test, test)


def infer_file_type(path):
    suffix = "".join(path.suffixes)

    if suffix in jsonl.JsonLinesFile.SUFFIXES:
        return jsonl.JsonLinesFile(path)

    if arrow is not None and suffix in arrow.File.SUFFIXES:
        return arrow.File.infer(path)

    return None


def _glob(path: Path, pattern="*", levels=1):
    if levels is not None:
        levels -= 1

    for subpath in path.glob(pattern):
        if subpath.is_dir():
            if levels != 0:
                yield from _glob(subpath, pattern, levels)
        else:
            yield subpath


def FileDataset(
    path: Path,
    freq: str,
    one_dim_target: bool = True,
    cache: bool = False,
    use_timestamp: bool = False,
    loader_class=None,
    ignore=False,
    pattern="*",
    levels=1,
) -> Dataset:
    path = Path(path)

    if not path.exists():
        raise FileNotFoundError(path)

    if path.is_dir():
        subpaths = _glob(path, pattern, levels)

        datasets = [
            FileDataset(
                subpath,
                freq,
                one_dim_target,
                cache,
                use_timestamp,
                loader_class=loader_class,
                ignore=True,
            )
            for subpath in subpaths
        ]

        return DatasetCollection(
            [dataset for dataset in datasets if dataset is not None]
        )

    assert path.is_file

    if loader_class is None:
        loader = infer_file_type(path)
        if loader is None:
            message = f"Cannot infer loader for {path}."
            if ignore:
                logging.warning(message)
                return None  # type: ignore

            raise ValueError(message)
    else:
        loader = loader_class(path)

    return _FileDataset(loader, freq, one_dim_target, cache, use_timestamp)


def _FileDataset(
    loader,
    freq: str,
    one_dim_target: bool = True,
    cache: bool = False,
    use_timestamp: bool = False,
) -> Dataset:
    process = ProcessDataEntry(
        freq, one_dim_target=one_dim_target, use_timestamp=use_timestamp
    )

    dataset: Dataset = Map(process, loader)

    if cache:
        dataset = Cached(dataset)

    return dataset


def ListDataset(
    data_iter: Iterable[DataEntry],
    freq: str,
    one_dim_target: bool = True,
    use_timestamp: bool = False,
) -> Dataset:
    """
    Dataset backed directly by a list of dictionaries.

    Parameters
    ----------
    data_iter
        Iterable object yielding all items in the dataset.
        Each item should be a dictionary mapping strings to values.
        For instance: {"start": "2014-09-07", "target": [0.1, 0.2]}.
    freq
        Frequency of the observation in the time series.
        Must be a valid Pandas frequency.
    one_dim_target
        Whether to accept only univariate target time series.
    """
    return Map(
        ProcessDataEntry(to_offset(freq), one_dim_target, use_timestamp),
        list(data_iter),
    )


@functools.lru_cache(10_000)
def _as_period(val, freq):
    return pd.Period(val, freq)


# TODO: find out whether this is a duplicate
class ProcessStartField(pydantic.BaseModel):
    """
    Transform the start field into a Period with the given frequency.

    Parameters
    ----------
    name
        Name of the field to transform.
    freq
        Frequency to use. This must be a valid Pandas frequency string.
    """

    class Config:
        arbitrary_types_allowed = True

    freq: Union[str, pd.DateOffset]
    use_timestamp: bool = False
    name: str = FieldName.START

    def __call__(self, data: DataEntry) -> DataEntry:
        try:
            if self.use_timestamp:
                data[self.name] = pd.Timestamp(data[self.name])
            else:
                data[self.name] = _as_period(data[self.name], self.freq)
        except (TypeError, ValueError) as e:
            raise GluonTSDataError(
                f'Error "{e}" occurred, when reading field "{self.name}"'
            ) from e

        return data


class ProcessTimeSeriesField:
    """
    Converts a time series field identified by `name` from a list of numbers
    into a numpy array.

    Constructor parameters modify the conversion logic in the following way:

    If `is_required=True`, throws a `GluonTSDataError` if the field is not
    present in the `Data` dictionary.

    If `is_cat=True`, the array type is `np.int32`, otherwise it is
    `np.float32`.

    If `is_static=True`, asserts that the resulting array is 1D,
    otherwise asserts that the resulting array is 2D. 2D dynamic arrays of
    shape (T) are automatically expanded to shape (1,T).

    Parameters
    ----------
    name
        Name of the field to process.
    is_required
        Whether the field must be present.
    is_cat
        Whether the field refers to categorical (i.e. integer) values.
    is_static
        Whether the field is supposed to have a time dimension.
    """

    # TODO: find a fast way to assert absence of nans.

    def __init__(
        self, name, is_required: bool, is_static: bool, is_cat: bool
    ) -> None:
        self.name = name
        self.is_required = is_required
        self.req_ndim = 1 if is_static else 2
        self.dtype = np.int32 if is_cat else np.float32

    def __call__(self, data: DataEntry) -> DataEntry:
        value = data.get(self.name, None)
        if value is not None:
            value = np.asarray(value, dtype=self.dtype)

            if self.req_ndim != value.ndim:
                raise GluonTSDataError(
                    f"Array '{self.name}' has bad shape - expected "
                    f"{self.req_ndim} dimensions, got {value.ndim}."
                )

            data[self.name] = value

            return data
        elif not self.is_required:
            return data
        else:
            raise GluonTSDataError(
                f"Object is missing a required field `{self.name}`"
            )


class ProcessDataEntry:
    def __init__(
        self,
        freq: str,
        one_dim_target: bool = True,
        use_timestamp: bool = False,
    ) -> None:
        # TODO: create a FormatDescriptor object that can be derived from a
        # TODO: Metadata and pass it instead of freq.
        # TODO: In addition to passing freq, the descriptor should be carry
        # TODO: information about required features.
        self.trans = cast(
            List[Callable[[DataEntry], DataEntry]],
            [
                ProcessStartField(freq=freq, use_timestamp=use_timestamp),
                # The next line abuses is_static=True in case of 1D targets.
                ProcessTimeSeriesField(
                    FieldName.TARGET,
                    is_required=True,
                    is_cat=False,
                    is_static=one_dim_target,
                ),
                ProcessTimeSeriesField(
                    FieldName.FEAT_DYNAMIC_CAT,
                    is_required=False,
                    is_cat=True,
                    is_static=False,
                ),
                ProcessTimeSeriesField(
                    FieldName.FEAT_DYNAMIC_REAL_LEGACY,  # backwards compatible
                    is_required=False,
                    is_cat=False,
                    is_static=False,
                ),
                ProcessTimeSeriesField(
                    FieldName.FEAT_DYNAMIC_REAL,
                    is_required=False,
                    is_cat=False,
                    is_static=False,
                ),
                ProcessTimeSeriesField(
                    FieldName.PAST_FEAT_DYNAMIC_REAL,
                    is_required=False,
                    is_cat=False,
                    is_static=False,
                ),
                ProcessTimeSeriesField(
                    FieldName.FEAT_STATIC_CAT,
                    is_required=False,
                    is_cat=True,
                    is_static=True,
                ),
                ProcessTimeSeriesField(
                    FieldName.FEAT_STATIC_REAL,
                    is_required=False,
                    is_cat=False,
                    is_static=True,
                ),
            ],
        )

    def __call__(self, data: DataEntry) -> DataEntry:
        for t in self.trans:
            data = t(data)
        return data


def load_datasets(
    metadata: Path,
    train: Path,
    test: Optional[Path],
    one_dim_target: bool = True,
    cache: bool = False,
) -> TrainDatasets:
    """
    Loads a dataset given metadata, train and test path.

    Parameters
    ----------
    metadata
        Path to the metadata file
    train
        Path to the training dataset files.
    test
        Path to the test dataset files.
    one_dim_target
        Whether to load FileDatasets as univariate target time series.
    cache
        Indicates whether the FileDatasets should be cached or not.

    Returns
    -------
    TrainDatasets
        An object collecting metadata, training data, test data.
    """
    meta = MetaData.parse_file(Path(metadata) / "metadata.json")
    train_ds = FileDataset(
        path=train, freq=meta.freq, one_dim_target=one_dim_target, cache=cache
    )
    test_ds = (
        FileDataset(
            path=test,
            freq=meta.freq,
            one_dim_target=one_dim_target,
            cache=cache,
        )
        if test
        else None
    )

    return TrainDatasets(metadata=meta, train=train_ds, test=test_ds)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from toolz.functoolz import compose_left

from gluonts.nursery.glide import Apply, ParApply
from gluonts.nursery.glide.util import lift


def lifted(fns):
    return tuple(map(lift, fns))


class Pipeline:
    def __init__(self, fns=()):
        self.fns = tuple(fns)

    def apply(self, parts, *args, **kwargs):
        return Apply(compose_left(*self.fns), parts, *args, **kwargs)

    def parapply(self, parts, *args, **kwargs):
        return ParApply(compose_left(*self.fns), parts, *args, **kwargs)

    def and_then(self, *fns):
        return Pipeline(self.fns + fns)

    def and_then_each(self, *fns):
        return self.and_then(*lifted(fns))

    def but_first(self, *fns):
        return Pipeline(fns + self.fns)

    def but_first_each(self, *fns):
        return self.but_first(*lifted(fns))

    def __add__(self, other):
        return Pipeline(self.fns + other.fns)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import inspect
import pydoc
from typing import Type, Union, cast

import pkg_resources
from toolz import keyfilter

from gluonts.model.estimator import Estimator
from gluonts.model.predictor import Predictor

from .exceptions import ForecasterNotFound

Forecaster = Type[Union[Estimator, Predictor]]


def forecaster_type_by_name(name: str) -> Forecaster:
    """
    Loads a forecaster from the `gluonts_forecasters` entry_points namespace by
    name.

    If a forecater wasn't register under that name, it tries to locate the
    class.

    Third-party libraries can register their forecasters as follows by defining
    a corresponding section in the `entry_points` section of their `setup.py`::

        entry_points={
            'gluonts_forecasters': [
                'model_a = my_models.model_a:MyEstimator',
                'model_b = my_models.model_b:MyPredictor',
            ]
        }
    """
    forecaster = None

    for entry_point in pkg_resources.iter_entry_points("gluonts_forecasters"):
        if entry_point.name == name:
            forecaster = entry_point.load()
            break
    else:
        forecaster = pydoc.locate(name)

    ForecasterNotFound.guard(
        forecaster is not None,
        f'Cannot locate estimator with classname "{name}".',
    )

    return cast(Forecaster, forecaster)


def invoke_with(fn, *args, **kwargs):
    """
    Call `fn(*args, **kwargs)`, but only use kwargs that `fn` actually uses.
    """

    # if `fn` has `**kwargs` argument, we can just call it directly
    if inspect.getfullargspec(fn).varkw is not None:
        return fn(*args, **kwargs)

    sig = inspect.signature(fn)

    kwargs = keyfilter(sig.parameters.__contains__, kwargs)

    arguments = sig.bind(*args, **kwargs).arguments
    return fn(**arguments)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from functools import partial
from typing import List, Optional

import numpy as np
from mxnet.gluon import HybridBlock
from pandas.tseries.frequencies import to_offset

from gluonts.core.component import validated
from gluonts.dataset.common import Dataset
from gluonts.dataset.field_names import FieldName
from gluonts.dataset.loader import (
    DataLoader,
    TrainDataLoader,
    ValidationDataLoader,
)
from gluonts.model.deepstate.issm import ISSM, CompositeISSM
from gluonts.model.predictor import Predictor
from gluonts.mx.batchify import batchify
from gluonts.mx.distribution.lds import ParameterBounds
from gluonts.mx.model.estimator import GluonEstimator
from gluonts.mx.model.predictor import RepresentableBlockPredictor
from gluonts.mx.trainer import Trainer
from gluonts.mx.util import copy_parameters, get_hybrid_forward_input_names
from gluonts.time_feature import (
    TimeFeature,
    norm_freq_str,
    time_features_from_frequency_str,
)
from gluonts.transform import (
    AddAgeFeature,
    AddObservedValuesIndicator,
    AddTimeFeatures,
    AsNumpyArray,
    CanonicalInstanceSplitter,
    Chain,
    ExpandDimArray,
    RemoveFields,
    SelectFields,
    SetField,
    TestSplitSampler,
    Transformation,
    VstackFeatures,
)

from ._network import DeepStatePredictionNetwork, DeepStateTrainingNetwork

SEASON_INDICATORS_FIELD = "seasonal_indicators"


# A dictionary mapping granularity to the period length of the longest season
# one can expect given the granularity of the time series. This is similar to
# the frequency value in the R forecast package:
# https://stats.stackexchange.com/questions/120806/frequency-value-for-seconds-minutes-intervals-data-in-r
# This is useful for setting default values for past/context length for models
# that do not do data augmentation and uses a single training example per time
# series in the dataset.
FREQ_LONGEST_PERIOD_DICT = {
    "M": 12,  # yearly seasonality
    "W": 52,  # yearly seasonality
    "D": 31,  # monthly seasonality
    "B": 22,  # monthly seasonality
    "H": 168,  # weekly seasonality
    "T": 1440,  # daily seasonality
}


def longest_period_from_frequency_str(freq_str: str) -> int:
    offset = to_offset(freq_str)
    return FREQ_LONGEST_PERIOD_DICT[norm_freq_str(offset.name)] // offset.n


class DeepStateEstimator(GluonEstimator):
    """
    Construct a DeepState estimator.

    This implements the deep state space model described in
    [RSG+18]_.

    Parameters
    ----------
    freq
        Frequency of the data to train on and predict
    prediction_length
        Length of the prediction horizon
    cardinality
        Number of values of each categorical feature.
        This must be set by default unless ``use_feat_static_cat``
        is set to `False` explicitly (which is NOT recommended).
    add_trend
        Flag to indicate whether to include trend component in the
        state space model
    past_length
        This is the length of the training time series;
        i.e., number of steps to unroll the RNN for before computing
        predictions.
        Set this to (at most) the length of the shortest time series in the
        dataset.
        (default: None, in which case the training length is set such that
        at least
        `num_seasons_to_train` seasons are included in the training.
        See `num_seasons_to_train`)
    num_periods_to_train
        (Used only when `past_length` is not set)
        Number of periods to include in the training time series. (default: 4)
        Here period corresponds to the longest cycle one can expect given
        the granularity of the time series.
        See: https://stats.stackexchange.com/questions/120806/frequency
        -value-for-seconds-minutes-intervals-data-in-r
    trainer
        Trainer object to be used (default: Trainer())
    num_layers
        Number of RNN layers (default: 2)
    num_cells
        Number of RNN cells for each layer (default: 40)
    cell_type
        Type of recurrent cells to use (available: 'lstm' or 'gru';
        default: 'lstm')
    num_parallel_samples
        Number of evaluation samples per time series to increase parallelism
        during inference.
        This is a model optimization that does not affect the accuracy (
        default: 100).
    dropout_rate
        Dropout regularization parameter (default: 0.1)
    use_feat_dynamic_real
        Whether to use the ``feat_dynamic_real`` field from the data
        (default: False)
    use_feat_static_cat
        Whether to use the ``feat_static_cat`` field from the data
        (default: True)
    embedding_dimension
        Dimension of the embeddings for categorical features
        (default: [min(50, (cat+1)//2) for cat in cardinality])
    scaling
        Whether to automatically scale the target values (default: true)
    time_features
        Time features to use as inputs of the RNN (default: None, in which
        case these are automatically determined based on freq)
    noise_std_bounds
        Lower and upper bounds for the standard deviation of the observation
        noise
    prior_cov_bounds
        Lower and upper bounds for the diagonal of the prior covariance matrix
    innovation_bounds
        Lower and upper bounds for the standard deviation of the observation
        noise
    batch_size
        The size of the batches to be used training and prediction.
    """

    @validated()
    def __init__(
        self,
        freq: str,
        prediction_length: int,
        cardinality: List[int],
        add_trend: bool = False,
        past_length: Optional[int] = None,
        num_periods_to_train: int = 4,
        trainer: Trainer = Trainer(
            epochs=100, num_batches_per_epoch=50, hybridize=False
        ),
        num_layers: int = 2,
        num_cells: int = 40,
        cell_type: str = "lstm",
        num_parallel_samples: int = 100,
        dropout_rate: float = 0.1,
        use_feat_dynamic_real: bool = False,
        use_feat_static_cat: bool = True,
        embedding_dimension: Optional[List[int]] = None,
        issm: Optional[ISSM] = None,
        scaling: bool = True,
        time_features: Optional[List[TimeFeature]] = None,
        noise_std_bounds: ParameterBounds = ParameterBounds(1e-6, 1.0),
        prior_cov_bounds: ParameterBounds = ParameterBounds(1e-6, 1.0),
        innovation_bounds: ParameterBounds = ParameterBounds(1e-6, 0.01),
        batch_size: int = 32,
    ) -> None:
        super().__init__(trainer=trainer, batch_size=batch_size)

        assert (
            prediction_length > 0
        ), "The value of `prediction_length` should be > 0"
        assert (
            past_length is None or past_length > 0
        ), "The value of `past_length` should be > 0"
        assert num_layers > 0, "The value of `num_layers` should be > 0"
        assert num_cells > 0, "The value of `num_cells` should be > 0"
        assert (
            num_parallel_samples > 0
        ), "The value of `num_parallel_samples` should be > 0"
        assert dropout_rate >= 0, "The value of `dropout_rate` should be >= 0"
        assert not use_feat_static_cat or any(c > 1 for c in cardinality), (
            "Cardinality of at least one static categorical feature must be"
            " larger than 1 if `use_feat_static_cat=True`. But cardinality"
            f" provided is: {cardinality}"
        )
        assert embedding_dimension is None or all(
            e > 0 for e in embedding_dimension
        ), "Elements of `embedding_dimension` should be > 0"

        assert all(
            np.isfinite(p.lower) and np.isfinite(p.upper) and p.lower > 0
            for p in [noise_std_bounds, prior_cov_bounds, innovation_bounds]
        ), (
            "All parameter bounds should be finite, and lower bounds should be"
            " positive"
        )

        self.past_length = (
            past_length
            if past_length is not None
            else num_periods_to_train * longest_period_from_frequency_str(freq)
        )
        self.prediction_length = prediction_length
        self.add_trend = add_trend
        self.num_layers = num_layers
        self.num_cells = num_cells
        self.cell_type = cell_type
        self.num_parallel_samples = num_parallel_samples
        self.scaling = scaling
        self.dropout_rate = dropout_rate
        self.use_feat_dynamic_real = use_feat_dynamic_real
        self.use_feat_static_cat = use_feat_static_cat
        self.cardinality = (
            cardinality if cardinality and use_feat_static_cat else [1]
        )
        self.embedding_dimension = (
            embedding_dimension
            if embedding_dimension is not None
            else [min(50, (cat + 1) // 2) for cat in self.cardinality]
        )

        self.issm = (
            issm
            if issm is not None
            else CompositeISSM.get_from_freq(freq, add_trend)
        )

        self.time_features = (
            time_features
            if time_features is not None
            else time_features_from_frequency_str(freq)
        )

        self.noise_std_bounds = noise_std_bounds
        self.prior_cov_bounds = prior_cov_bounds
        self.innovation_bounds = innovation_bounds

    def create_transformation(self) -> Transformation:
        remove_field_names = [
            FieldName.FEAT_DYNAMIC_CAT,
            FieldName.FEAT_STATIC_REAL,
        ]
        if not self.use_feat_dynamic_real:
            remove_field_names.append(FieldName.FEAT_DYNAMIC_REAL)

        return Chain(
            [RemoveFields(field_names=remove_field_names)]
            + (
                [SetField(output_field=FieldName.FEAT_STATIC_CAT, value=[0.0])]
                if not self.use_feat_static_cat
                else []
            )
            + [
                AsNumpyArray(field=FieldName.FEAT_STATIC_CAT, expected_ndim=1),
                AsNumpyArray(field=FieldName.TARGET, expected_ndim=1),
                # gives target the (1, T) layout
                ExpandDimArray(field=FieldName.TARGET, axis=0),
                AddObservedValuesIndicator(
                    target_field=FieldName.TARGET,
                    output_field=FieldName.OBSERVED_VALUES,
                ),
                # Unnormalized seasonal features
                AddTimeFeatures(
                    time_features=self.issm.time_features(),
                    pred_length=self.prediction_length,
                    start_field=FieldName.START,
                    target_field=FieldName.TARGET,
                    output_field=SEASON_INDICATORS_FIELD,
                ),
                AddTimeFeatures(
                    start_field=FieldName.START,
                    target_field=FieldName.TARGET,
                    output_field=FieldName.FEAT_TIME,
                    time_features=self.time_features,
                    pred_length=self.prediction_length,
                ),
                AddAgeFeature(
                    target_field=FieldName.TARGET,
                    output_field=FieldName.FEAT_AGE,
                    pred_length=self.prediction_length,
                    log_scale=True,
                ),
                VstackFeatures(
                    output_field=FieldName.FEAT_TIME,
                    input_fields=[FieldName.FEAT_TIME, FieldName.FEAT_AGE]
                    + (
                        [FieldName.FEAT_DYNAMIC_REAL]
                        if self.use_feat_dynamic_real
                        else []
                    ),
                ),
            ]
        )

    def _create_instance_splitter(self, mode: str):
        assert mode in ["training", "validation", "test"]

        return CanonicalInstanceSplitter(
            target_field=FieldName.TARGET,
            is_pad_field=FieldName.IS_PAD,
            start_field=FieldName.START,
            forecast_start_field=FieldName.FORECAST_START,
            instance_sampler=TestSplitSampler(),
            time_series_fields=[
                FieldName.FEAT_TIME,
                SEASON_INDICATORS_FIELD,
                FieldName.OBSERVED_VALUES,
            ],
            allow_target_padding=True,
            instance_length=self.past_length,
            use_prediction_features=(mode != "training"),
            prediction_length=self.prediction_length,
        )

    def create_training_data_loader(
        self,
        data: Dataset,
        **kwargs,
    ) -> DataLoader:
        input_names = get_hybrid_forward_input_names(DeepStateTrainingNetwork)
        instance_splitter = self._create_instance_splitter("training")
        return TrainDataLoader(
            dataset=data,
            transform=instance_splitter + SelectFields(input_names),
            batch_size=self.batch_size,
            stack_fn=partial(batchify, ctx=self.trainer.ctx, dtype=self.dtype),
            **kwargs,
        )

    def create_validation_data_loader(
        self,
        data: Dataset,
        **kwargs,
    ) -> DataLoader:
        input_names = get_hybrid_forward_input_names(DeepStateTrainingNetwork)
        instance_splitter = self._create_instance_splitter("validation")
        return ValidationDataLoader(
            dataset=data,
            transform=instance_splitter + SelectFields(input_names),
            batch_size=self.batch_size,
            stack_fn=partial(batchify, ctx=self.trainer.ctx, dtype=self.dtype),
        )

    def create_training_network(self) -> DeepStateTrainingNetwork:
        return DeepStateTrainingNetwork(
            num_layers=self.num_layers,
            num_cells=self.num_cells,
            cell_type=self.cell_type,
            past_length=self.past_length,
            prediction_length=self.prediction_length,
            issm=self.issm,
            dropout_rate=self.dropout_rate,
            cardinality=self.cardinality,
            embedding_dimension=self.embedding_dimension,
            scaling=self.scaling,
            noise_std_bounds=self.noise_std_bounds,
            prior_cov_bounds=self.prior_cov_bounds,
            innovation_bounds=self.innovation_bounds,
        )

    def create_predictor(
        self, transformation: Transformation, trained_network: HybridBlock
    ) -> Predictor:
        prediction_splitter = self._create_instance_splitter("test")

        prediction_network = DeepStatePredictionNetwork(
            num_layers=self.num_layers,
            num_cells=self.num_cells,
            cell_type=self.cell_type,
            past_length=self.past_length,
            prediction_length=self.prediction_length,
            issm=self.issm,
            dropout_rate=self.dropout_rate,
            cardinality=self.cardinality,
            embedding_dimension=self.embedding_dimension,
            scaling=self.scaling,
            num_parallel_samples=self.num_parallel_samples,
            noise_std_bounds=self.noise_std_bounds,
            prior_cov_bounds=self.prior_cov_bounds,
            innovation_bounds=self.innovation_bounds,
            params=trained_network.collect_params(),
        )

        copy_parameters(trained_network, prediction_network)

        return RepresentableBlockPredictor(
            input_transform=transformation + prediction_splitter,
            prediction_net=prediction_network,
            batch_size=self.batch_size,
            prediction_length=self.prediction_length,
            ctx=self.trainer.ctx,
        )



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import itertools
import logging
import time
from typing import List, Optional
import mxnet as mx
import numpy as np
from gluonts.core.component import validated
from gluonts.dataset.loader import DataLoader
from gluonts.gluonts_tqdm import tqdm
from gluonts.mx.trainer import Trainer
from gluonts.mx.util import HybridContext
from mxnet import autograd
from mxnet.gluon import nn
from mxnet.metric import ndarray
from .callbacks import Callback, CallbackList

logger = logging.getLogger("gluonts_meta").getChild("trainer")

# make the IDE happy: mx.py does not explicitly import autograd
mx.autograd = autograd  # type: ignore


class TimedTrainer(Trainer):
    """
    A custom trainer whose training duration is based on wall clock time
    instead of epochs.
    """

    @validated()
    def __init__(
        self,
        training_time: float,
        validation_milestones: Optional[List[float]] = None,
        learning_rate: float = 1e-3,
        callbacks: Optional[List[Callback]] = None,
    ) -> None:
        super().__init__(learning_rate=learning_rate)

        validation_milestones = validation_milestones or []
        assert all(
            x < y
            for x, y in zip(validation_milestones, validation_milestones[1:])
        ), "Validation milestones must be increasing."

        self.training_time = training_time
        self.validation_milestones = validation_milestones or []
        self.callbacks = CallbackList(callbacks or [])

    def __call__(
        self,
        net: nn.HybridBlock,
        train_iter: DataLoader,
        validation_iter: Optional[DataLoader] = None,
    ) -> None:
        logger.info("Start model training")
        net.initialize(ctx=self.ctx, init=self.init)

        with HybridContext(
            net=net,
            hybridize=self.hybridize,
            static_alloc=True,
            static_shape=True,
        ):
            self._train_loop(net, train_iter, validation_iter)

    def _train_loop(  # pylint: disable=too-many-statements
        self,
        net: nn.HybridBlock,
        train_iter: DataLoader,
        validation_iter: Optional[DataLoader],
    ) -> None:
        optimizer = mx.optimizer.Adam(
            learning_rate=self.learning_rate,
            wd=self.weight_decay,
            clip_gradient=self.clip_gradient,
        )

        trainer = mx.gluon.Trainer(
            net.collect_params(),
            optimizer=optimizer,
            kvstore="device",
        )

        first_forward = True
        time_elapsed = 0
        validation_idx = 0

        def loop(
            batch_iter: DataLoader,
            num_batches_to_use: Optional[int] = None,
            is_training: bool = True,
        ) -> mx.metric.Loss:
            nonlocal first_forward, time_elapsed, validation_idx

            tic = time.time()
            subtic = 0

            epoch_loss = mx.metric.Loss()
            batch_iter = itertools.islice(batch_iter, num_batches_to_use)

            it = tqdm(batch_iter, total=num_batches_to_use)
            for batch_no, batch in enumerate(it, start=1):
                # `batch` here is expected to be a dictionary whose fields
                # should correspond 1-to-1 with the network inputs
                # see below how `batch.values()` is fed into the network
                if first_forward:
                    tictic = time.time()
                    first_forward = False
                    _ = net(*batch.values())
                    self.callbacks.on_network_initialization_end(net)
                    subtic += time.time() - tictic

                with mx.autograd.record():  # type: ignore
                    # we set the mode explicitly as by default mxnet assumes
                    # predict mode and hence dropout layers are not used if
                    # the mode is not explicitly set to training
                    mode = (
                        autograd.train_mode
                        if is_training
                        else autograd.predict_mode
                    )
                    with mode():
                        output = net(*batch.values())

                    # network can returns several outputs, the first being
                    # always the loss when having multiple outputs, the
                    # forward returns a list in the case of hybrid and a
                    # tuple otherwise we may wrap network outputs in the
                    # future to avoid this type check
                    if isinstance(output, (list, tuple)):
                        loss = output[0]
                    else:
                        loss = output

                    batch_size = loss.shape[0]

                # pylint: disable=no-member
                if not np.isfinite(ndarray.sum(loss).asscalar()):  # type: ignore
                    logger.warning(
                        "Batch [%d] gave NaN loss and it will be ignored",
                        batch_no,
                    )
                else:
                    if is_training:
                        loss.backward()
                        trainer.step(batch_size)
                    epoch_loss.update(None, preds=loss)

                if is_training:
                    total_time_elapsed = (
                        time_elapsed + time.time() - tic - subtic
                    )

                    orig_lr = trainer.learning_rate
                    tictic = time.time()
                    self.callbacks.on_train_batch_end(net, total_time_elapsed)
                    subtic += time.time() - tictic
                    if trainer.learning_rate != orig_lr:
                        logger.info(
                            "Trainer learning rate set to %f",
                            trainer.learning_rate,
                        )

                lv = _loss_value(epoch_loss)
                it.set_postfix(
                    ordered_dict={
                        ("" if is_training else "validation_")
                        + "avg_epoch_loss": lv,
                    },
                    refresh=False,
                )

                # Check if should finish
                if is_training:
                    if total_time_elapsed > self.training_time:  # type: ignore
                        time_elapsed = total_time_elapsed  # type: ignore
                        break
                    if len(self.validation_milestones) > validation_idx and (
                        total_time_elapsed  # type: ignore
                        > self.validation_milestones[validation_idx]
                    ):
                        time_elapsed = total_time_elapsed  # type: ignore
                        validation_idx += 1
                        break
                # If validating, call the callback with the loss
                else:
                    self.callbacks.on_validation_epoch_end(lv)

            # mark epoch end time and log time cost of current epoch
            toc = time.time()
            logger.info("Elapsed time %.3f seconds", toc - tic)
            logger.info(
                "Evaluation metric '%s'=%f",
                ("" if is_training else "validation_") + "epoch_loss",
                lv,  # type: ignore
            )

            return epoch_loss

        self.callbacks.on_train_start(trainer)
        while True:
            loop(train_iter)
            if validation_iter is not None:
                loop(validation_iter, is_training=False)
            if time_elapsed > self.training_time:
                break

        logger.info("End model training")


def _loss_value(loss: mx.metric.Loss) -> float:
    return loss.get_name_value()[0][1]



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from abc import ABC
from typing import cast, Generic, List, Optional, TypeVar
from tsbench.config import Config, DatasetConfig, EnsembleConfig, ModelConfig
from tsbench.evaluations.metrics import Performance
from ._recommendation import Recommendation
from .generator import CandidateGenerator, ReplayCandidateGenerator
from .utils import argsort_nondominated

T = TypeVar("T", ModelConfig, EnsembleConfig)


class Recommender(ABC, Generic[T]):
    """
    A recommender uses a surrogate to recommend models and their configurations
    based on desired target metrics.

    This class implements the general interface.
    """

    def __init__(
        self,
        objectives: List[str],
        focus: Optional[str] = None,
        generator: Optional[CandidateGenerator[T]] = None,
    ):
        """
        Args:
            objectives: The list of performance metrics to minimize.
            focus: The metric to prefer. Must be either in the list of objectives. If not
                provided, the first metric to optimize is chosen.
            generator: The generator that generates configurations for recommendations. By default,
                this is the replay candidate generator.
        """
        # Assertions
        assert len(objectives) > 0, "No metrics provided."

        assert (
            focus is None or focus in objectives
        ), "Focus metric not found in metrics to optimize."

        # Initialize attributes
        self.generator = generator or ReplayCandidateGenerator()
        self.objectives = objectives
        self.focus = focus

    @property
    def required_cpus(self) -> int:
        """
        The number of CPUs required for fitting the recommender.
        """
        return 1

    @property
    def required_memory(self) -> int:
        """
        The amount of memory in GiB required for fitting the recommender.
        """
        return 1

    def fit(
        self, configs: List[Config[T]], _performances: List[Performance]
    ) -> None:
        """
        Fits the recommender, including surrogate model and generator, on the
        provided configurations.

        Args:
            configs: The configurations to train on (the generator typically extracts the unique
                model configurations).
            performances: The performances that the surrogate should fit on. The performances must
                align with the provided configurations.
        """
        self.generator.fit(list({c.model for c in configs}))

    def recommend(
        self,
        dataset: DatasetConfig,
        candidates: Optional[List[T]] = None,
        max_count: int = 10,
    ) -> List[Recommendation[T]]:
        """
        This method takes a dataset and a set of constraints and outputs a set
        of recommendations. The recommendations provide both the configurations
        of the recommended model as well as the expected performance.

        Args:
            dataset: The configuration of the dataset for which to recommend a model.
            candidates: A list of model configurations that are allowed to be recommended. If
                `None`, any model configuration is permitted.
            max_count: The maximum number of models to recommend.

        Returns:
            The recommendations which (approximately) satisfy the provided constraints.
        """
        model_configs = self.generator.generate(candidates)
        configs = [Config(m, dataset) for m in model_configs]
        performances = self._get_performances(configs)

        # We construct a data frame, extracting the performance metrics to minimize.
        # Then, we invert the performance metrics for the metrics to maximize.
        df = Performance.to_dataframe(performances)[self.objectives]

        # Then, we perform a nondominated sort
        argsort = argsort_nondominated(
            df.to_numpy(),  # type: ignore
            dim=df.columns.tolist().index(self.focus)
            if self.focus is not None
            else None,
            max_items=max_count,
        )

        # And get the recommendations
        result = []
        for choice in cast(List[int], argsort):
            config = configs[choice]
            recommendation = Recommendation(config.model, performances[choice])
            result.append(recommendation)

        return result

    def _get_performances(self, configs: List[Config[T]]) -> List[Performance]:
        raise NotImplementedError



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import cast, List, Literal, Optional
import numpy as np
import numpy.typing as npt
import pytorch_lightning as pl
import torch
from lightkit.data import DataLoader
from torch import nn
from torch.utils.data import TensorDataset
from tsbench.config import Config, ModelConfig
from tsbench.evaluations.tracking import ModelTracker
from ._base import DatasetFeaturesMixin, OutputNormalization, Surrogate
from ._factory import register_surrogate
from .torch import ListMLELoss, MLPLightningModule
from .transformers import ConfigTransformer


@register_surrogate("mlp")
class MLPSurrogate(Surrogate[ModelConfig], DatasetFeaturesMixin):
    """
    The MLP surrogate predicts a model's performance on a new dataset using an
    MLP.

    The MLP converts inputs into feature vectors of the same size and uses
    either ranking or regression to predict metrics.
    """

    trainer_: pl.Trainer
    models_: List[nn.Module]

    def __init__(
        self,
        tracker: ModelTracker,
        objective: Literal["regression", "ranking"] = "regression",
        discount: Optional[
            Literal["logarithmic", "linear", "quadratic"]
        ] = None,
        hidden_layer_sizes: Optional[List[int]] = None,
        weight_decay: float = 0.0,
        dropout: float = 0.0,
        use_simple_dataset_features: bool = False,
        use_seasonal_naive_performance: bool = False,
        use_catch22_features: bool = False,
        predict: Optional[List[str]] = None,
        output_normalization: OutputNormalization = None,
        impute_simulatable: bool = False,
    ):
        """
        Args:
            tracker: A tracker that can be used to impute latency and number of model parameters
                into model performances. Also, it is required for some input features.
            objective: The optimization objective for the XGBoost estimators.
            discount: The discount to apply for the ranking loss. If provided, it focuses on
                correctly predicting the top values.
            hidden_layer_sizes: The dimensions of the hidden layers. Defaults to no hidden layers,
                i.e. a linear predictor.
            weight_decay: The weight decay to apply during optimization.
            dropout: The dropout probability of dropout layers applied after every activation
                function.
            use_simple_dataset_features: Whether to use dataset features to predict using a
                weighted average.
            use_seasonal_naive_performance: Whether to use the Seasonal Naïve nCRPS as dataset
                featuers. Requires the cacher to be set.
            use_catch22_features: Whether to use catch22 features for datasets statistics. Ignored
                if `use_dataset_features` is not set.
            predict: The metrics to predict. All if not provided.
            output_normalization: The type of normalization to apply to the features of each
                dataset independently. `None` applies no normalization, "quantile" applies quantile
                normalization, and "standard" transforms data to have zero mean and unit variance.
            impute_simulatable: Whether the tracker should impute latency and number of model
                parameters into the returned performance object.
        """
        super().__init__(
            tracker, predict, output_normalization, impute_simulatable
        )

        self.use_ranking = objective == "ranking"
        self.hidden_layer_sizes = hidden_layer_sizes or []
        self.weight_decay = weight_decay
        self.dropout = dropout

        self.config_transformer = ConfigTransformer(
            add_model_features=True,
            add_dataset_statistics=use_simple_dataset_features,
            add_seasonal_naive_performance=use_seasonal_naive_performance,
            add_catch22_features=use_catch22_features,
            tracker=tracker,
        )

        if objective == "regression":
            self.loss = nn.MSELoss()
        elif objective == "ranking":
            self.loss = ListMLELoss(discount=discount)

    @property
    def required_cpus(self) -> int:
        return 4

    def _fit(
        self, X: List[Config[ModelConfig]], y: npt.NDArray[np.float32]
    ) -> None:
        # Fit transformers to infer dimensionality
        X_numpy = self.config_transformer.fit_transform(X)

        input_dim = len(self.config_transformer.feature_names_)
        output_dim = y.shape[1]

        # For initializing data, we prepare group IDs for the datasets
        mapping = {d: i for i, d in enumerate({x.dataset for x in X})}

        # For each output variable, we need to train a separate model
        self.models_ = []
        for i in range(output_dim):
            model = self._init_model(input_dim)
            module = MLPLightningModule(model, self.loss, self.weight_decay)

            # Train on output variable i
            dataset = TensorDataset(
                torch.from_numpy(X_numpy).float(),
                torch.from_numpy(y[:, i : i + 1]).float(),
                torch.as_tensor(
                    [mapping[x.dataset] for x in X], dtype=torch.long
                ),
            )
            train_loader = DataLoader(dataset, batch_size=len(dataset))
            self._trainer.fit(module, train_dataloaders=train_loader)

            # Add to models
            self.models_.append(model)

    def _predict(
        self, X: List[Config[ModelConfig]]
    ) -> npt.NDArray[np.float32]:
        # Get data
        X_numpy = self.config_transformer.transform(X)
        dataset = TensorDataset(
            torch.from_numpy(X_numpy).float(),
            torch.zeros(len(X_numpy)),  # dummy data due to PL bug
        )
        test_loader = DataLoader(dataset, batch_size=len(dataset))

        # Run prediction
        predictions = []
        for model in self.models_:
            module = MLPLightningModule(model, self.loss)
            out = cast(
                List[torch.Tensor], self._trainer.predict(module, test_loader)
            )
            predictions.append(out[0].numpy())

        return np.concatenate(predictions, axis=-1)

    @property
    def _trainer(self) -> pl.Trainer:
        return pl.Trainer(
            max_epochs=1000,
            logger=False,
            enable_checkpointing=False,
            enable_model_summary=False,
            enable_progress_bar=False,
            gpus=int(torch.cuda.is_available()),
        )

    def _init_model(self, input_dim: int) -> nn.Module:
        layer_sizes = [input_dim] + self.hidden_layer_sizes + [1]
        layers = []
        for i, (in_size, out_size) in enumerate(
            zip(layer_sizes, layer_sizes[1:])
        ):
            if i > 0:
                layers.append(nn.LeakyReLU())
                if self.dropout > 0:
                    layers.append(nn.Dropout(self.dropout))
            layers.append(nn.Linear(in_size, out_size))
        return nn.Sequential(*layers)



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

import logging
from functools import singledispatch
from typing import Callable, Iterator, List, Optional

import numpy as np

from gluonts.core.component import validated
from gluonts.dataset.common import DataEntry
from gluonts.dataset.field_names import FieldName
from gluonts.dataset.loader import DataLoader
from gluonts.model.forecast import Forecast, QuantileForecast, SampleForecast

logger = logging.getLogger(__name__)

OutputTransform = Callable[[DataEntry, np.ndarray], np.ndarray]

LOG_CACHE = set()
OUTPUT_TRANSFORM_NOT_SUPPORTED_MSG = (
    "The `output_transform` argument is not supported and will be ignored."
)
NOT_SAMPLE_BASED_MSG = (
    "Forecast is not sample based. Ignoring parameter `num_samples` from"
    " predict method."
)


def log_once(msg):
    global LOG_CACHE
    if msg not in LOG_CACHE:
        logger.info(msg)
        LOG_CACHE.add(msg)


# different deep learning frameworks generate predictions and the tensor to
# numpy conversion differently, use a dispatching function to prevent needing
# a ForecastGenerators for each framework
@singledispatch
def predict_to_numpy(prediction_net, args) -> np.ndarray:
    raise NotImplementedError


@singledispatch
def recursively_zip_arrays(x) -> Iterator:
    """
    Helper function to recursively zip nested collections of arrays.

    This defines the fallback implementation, which one can specialized for
    specific types using by doing ``@recursively_zip_arrays.register`` on the
    type. Implementations for lists, tuples, and NumPy arrays are provided.

    For an array `a` (e.g. a numpy array)

        _extract_instances(a) -> [a[0], a[1], ...]

    For (nested) tuples of arrays `(a, (b, c))`

        _extract_instances((a, (b, c)) -> [(a[0], (b[0], c[0])), (a[1], (b
         [1], c[1])), ...]
    """
    raise NotImplementedError


@recursively_zip_arrays.register(np.ndarray)
def _(x: np.ndarray) -> Iterator[list]:
    for i in range(x.shape[0]):
        yield x[i]


@recursively_zip_arrays.register(tuple)
def _(x: tuple) -> Iterator[tuple]:
    for m in zip(*[recursively_zip_arrays(y) for y in x]):
        yield tuple(r for r in m)


@recursively_zip_arrays.register(list)
def _(x: list) -> Iterator[list]:
    for m in zip(*[recursively_zip_arrays(y) for y in x]):
        yield [r for r in m]


@recursively_zip_arrays.register(type(None))
def _(x: type(None)) -> Iterator[type(None)]:
    while True:
        yield None


@singledispatch
def make_distribution_forecast(distr, *args, **kwargs) -> Forecast:
    raise NotImplementedError


class ForecastGenerator:
    """
    Classes used to bring the output of a network into a class.
    """

    def __call__(
        self,
        inference_data_loader: DataLoader,
        prediction_net,
        input_names: List[str],
        output_transform: Optional[OutputTransform],
        num_samples: Optional[int],
        **kwargs
    ) -> Iterator[Forecast]:
        raise NotImplementedError()


class QuantileForecastGenerator(ForecastGenerator):
    @validated()
    def __init__(self, quantiles: List[str]) -> None:
        self.quantiles = quantiles

    def __call__(
        self,
        inference_data_loader: DataLoader,
        prediction_net,
        input_names: List[str],
        output_transform: Optional[OutputTransform],
        num_samples: Optional[int],
        **kwargs
    ) -> Iterator[Forecast]:
        for batch in inference_data_loader:
            inputs = [batch[k] for k in input_names]
            outputs = predict_to_numpy(prediction_net, inputs)
            if output_transform is not None:
                outputs = output_transform(batch, outputs)

            if num_samples:
                log_once(NOT_SAMPLE_BASED_MSG)

            i = -1
            for i, output in enumerate(outputs):
                yield QuantileForecast(
                    output,
                    start_date=batch[FieldName.FORECAST_START][i],
                    item_id=batch[FieldName.ITEM_ID][i]
                    if FieldName.ITEM_ID in batch
                    else None,
                    info=batch["info"][i] if "info" in batch else None,
                    forecast_keys=self.quantiles,
                )
            assert i + 1 == len(batch[FieldName.FORECAST_START])


class SampleForecastGenerator(ForecastGenerator):
    @validated()
    def __init__(self):
        pass

    def __call__(
        self,
        inference_data_loader: DataLoader,
        prediction_net,
        input_names: List[str],
        output_transform: Optional[OutputTransform],
        num_samples: Optional[int],
        **kwargs
    ) -> Iterator[Forecast]:
        for batch in inference_data_loader:
            inputs = [batch[k] for k in input_names]
            outputs = predict_to_numpy(prediction_net, inputs)
            if output_transform is not None:
                outputs = output_transform(batch, outputs)
            if num_samples:
                num_collected_samples = outputs[0].shape[0]
                collected_samples = [outputs]
                while num_collected_samples < num_samples:
                    outputs = predict_to_numpy(prediction_net, inputs)
                    if output_transform is not None:
                        outputs = output_transform(batch, outputs)
                    collected_samples.append(outputs)
                    num_collected_samples += outputs[0].shape[0]
                outputs = [
                    np.concatenate(s)[:num_samples]
                    for s in zip(*collected_samples)
                ]
                assert len(outputs[0]) == num_samples
            i = -1
            for i, output in enumerate(outputs):
                yield SampleForecast(
                    output,
                    start_date=batch[FieldName.FORECAST_START][i],
                    item_id=batch[FieldName.ITEM_ID][i]
                    if FieldName.ITEM_ID in batch
                    else None,
                    info=batch["info"][i] if "info" in batch else None,
                )
            assert i + 1 == len(batch[FieldName.FORECAST_START])


class DistributionForecastGenerator(ForecastGenerator):
    @validated()
    def __init__(self, distr_output) -> None:
        self.distr_output = distr_output

    def __call__(
        self,
        inference_data_loader: DataLoader,
        prediction_net,
        input_names: List[str],
        output_transform: Optional[OutputTransform],
        num_samples: Optional[int],
        **kwargs
    ) -> Iterator[Forecast]:
        for batch in inference_data_loader:
            inputs = [batch[k] for k in input_names]
            outputs = prediction_net(*inputs)

            if output_transform:
                log_once(OUTPUT_TRANSFORM_NOT_SUPPORTED_MSG)
            if num_samples:
                log_once(NOT_SAMPLE_BASED_MSG)

            distributions = [
                self.distr_output.distribution(*u)
                for u in recursively_zip_arrays(outputs)
            ]

            i = -1
            for i, distr in enumerate(distributions):
                yield make_distribution_forecast(
                    distr,
                    start_date=batch[FieldName.FORECAST_START][i],
                    item_id=batch[FieldName.ITEM_ID][i]
                    if FieldName.ITEM_ID in batch
                    else None,
                    info=batch["info"][i] if "info" in batch else None,
                )
            assert i + 1 == len(batch[FieldName.FORECAST_START])



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from typing import List, Optional, Union
import numpy as np
import numpy.typing as npt
import scipy.stats as st


def pareto_efficiency_mask(X: npt.NDArray[np.float32]) -> np.ndarray:
    """
    Evaluates for each allocation in the provided array whether it is Pareto
    efficient. The costs are assumed to be improved by lowering them.

    Args:
        X: Array of shape [N, D] (N: number of allocations, D: number of costs) containing the
            allocations to check.

    Returns:
        A boolean array of shape [N], indicating for each allocation whether it is Pareto
            efficient.
    """
    # First, we assume that all allocations are Pareto efficient, i.e. not dominated
    mask = np.ones(X.shape[0], dtype=bool)
    # Then, we iterate over all allocations A and check which are dominated by then current
    # allocation A. If it is, we don't need to check it against another allocation.
    for i, allocation in enumerate(X):
        # Only consider allocation if it hasn't been dominated yet
        if mask[i]:
            # An allocation is dominated by A if all costs are equal or lower and at least one cost
            # is strictly lower. Using that definition, A cannot be dominated by itself.
            dominated = np.all(allocation <= X[mask], axis=1) * np.any(
                allocation < X[mask], axis=1
            )
            mask[mask] = ~dominated

    return mask


def epsilon_net_indices(
    X: npt.NDArray[np.float32], dim: Optional[int] = None
) -> np.ndarray:
    """
    Outputs an order of the items in the provided array such that the items are
    spaced well. This means that after choosing a seed item, the next item is
    chosen to be the farthest from the seed item. The third item is then chosen
    to maximize the distance to the existing points and so on.

    This algorithm is taken from "Nearest-Neighbor Searching and Metric Space Dimensions"
    (Clarkson, 2005, p.17).

    Args:
        X: Array of shape [N, D] (N: number of items, D: dimensionality) with the items to
            sparsify.
        dim: The index of the dimension which to use to choose the seed item. If `None`, an item is
            chosen at random, otherwise the item with the lowest value in the specified dimension
            is used.

    Returns:
        An array of shape [N] providing the item indices that define a sparsified order of the
            items.
    """
    indices = set(range(X.shape[0]))

    # Choose the seed item according to dim
    if dim is None:
        seed = np.random.choice(X.shape[0])
    else:
        seed = np.argmin(X, axis=0)[dim]

    # Initialize the order
    order = [seed]
    indices.remove(seed)

    # Iterate until all models have been chosen
    while indices:
        # Get the distance to all items that have already been chosen
        ordered_indices = list(indices)
        diff = (
            X[ordered_indices][:, None, :].repeat(len(order), axis=1)
            - X[order]
        )
        min_distances = np.linalg.norm(diff, axis=-1).min(-1)

        # Then, choose the one with the maximum distance to all points
        choice = ordered_indices[min_distances.argmax()]
        order.append(choice)
        indices.remove(choice)

    return np.array(order)


def argsort_nondominated(
    X: npt.NDArray[np.float32],
    dim: Optional[int] = None,
    max_items: Optional[int] = None,
    flatten: bool = True,
) -> Union[List[int], List[List[int]]]:
    """
    Performs a multi-objective sort by iteratively computing the Pareto front
    and sparsifying the items within the Pareto front. This is a non-dominated
    sort leveraging an epsilon-net. All metrics are normalized using quantile
    normalization.

    Args:
        X: Array of shape [N, D] (N: number of items, D: dimensionality) with the
            multi-dimensional items to sort.
        dim: The feature (metric) to prefer when ranking items within the Pareto front. If `None`,
            items are chosen randomly.
        max_items: The maximum number of items that should be returned. When this is `None`, all
            items are sorted.
        flatten: Whether to flatten the resulting array.

    Returns:
        The indices of the sorted items, either globally or within each of the Pareto fronts,
        depending on the value of `flatten`.
    """
    remaining = np.arange(X.shape[0])
    indices = []
    num_items = 0

    # First, we want to normalize the input by performing a quantile transform into the range
    # [0, 1]
    X = (st.rankdata(X, axis=0) - 1) / (X.shape[0] - 1)  # type: ignore

    # Iterate until max_items are reached or there are no items left
    while remaining.size > 0 and (max_items is None or num_items < max_items):
        # Compute the Pareto front and sort the items within
        pareto_mask = pareto_efficiency_mask(X[remaining])
        pareto_front = remaining[pareto_mask]
        pareto_order = epsilon_net_indices(X[pareto_front], dim=dim)

        # Add order to the indices
        indices.append(pareto_front[pareto_order].tolist())
        num_items += len(pareto_front)

        # Remove items in the Pareto front from the remaining items
        remaining = remaining[~pareto_mask]

    # Restrict the number of items returned and optionally flatten
    if max_items is not None:
        limit = max_items - sum(len(x) for x in indices[:-1])
        indices[-1] = indices[-1][:limit]
        if not indices[-1]:
            indices = indices[:-1]

    if flatten:
        return [i for ix in indices for i in ix]
    return indices



# Copyright 2018 Amazon.com, Inc. or its affiliates. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License").
# You may not use this file except in compliance with the License.
# A copy of the License is located at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# or in the "license" file accompanying this file. This file is distributed
# on an "AS IS" BASIS, WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either
# express or implied. See the License for the specific language governing
# permissions and limitations under the License.

from .Adam import Adam
from .Adagrad import Adagrad
from .SAdagrad import SAdagrad
from .SAdam import SAdam
from .SCott import SCott
from .SCSG import SCSG
from .SGD import SGD
